"""utils.py - Helper functions for building the model and for loading model parameters.
   These helper functions are built to mirror those in the official TensorFlow implementation.
"""

# Author: lukemelas (github username)
# Github repo: https://github.com/lukemelas/EfficientNet-PyTorch
# With adjustments and added comments by workingcoder (github username).

import re
import math
import collections
from functools import partial
import torch
from torch import nn
from torch.nn import functional as F
from torch.utils import model_zoo


################################################################################
### Help functions for model architecture
################################################################################

# GlobalParams and BlockArgs: Two namedtuples
# Swish and MemoryEfficientSwish: Two implementations of the method
# round_filters and round_repeats:
#     Functions to calculate params for scaling model width and depth ! ! !
# get_width_and_height_from_size and calculate_output_image_size
# drop_connect: A structural design
# get_same_padding_conv2d:
#     Conv2dDynamicSamePadding
#     Conv2dStaticSamePadding
# get_same_padding_maxPool2d:
#     MaxPool2dDynamicSamePadding
#     MaxPool2dStaticSamePadding
#     It's an additional function, not used in EfficientNet,
#     but can be used in other model (such as EfficientDet).

# Parameters for the entire model (stem, all blocks, and head)
GlobalParams = collections.namedtuple('GlobalParams', [
    'width_coefficient', 'depth_coefficient', 'image_size', 'dropout_rate',
    'num_classes', 'batch_norm_momentum', 'batch_norm_epsilon',
    'drop_connect_rate', 'depth_divisor', 'min_depth', 'include_top'])

# Parameters for an individual model block
BlockArgs = collections.namedtuple('BlockArgs', [
    'num_repeat', 'kernel_size', 'stride', 'expand_ratio',
    'input_filters', 'output_filters', 'se_ratio', 'id_skip'])

# Set GlobalParams and BlockArgs's defaults
GlobalParams.__new__.__defaults__ = (None,) * len(GlobalParams._fields)
BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)


# An ordinary implementation of Swish function
class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


# A memory-efficient implementation of Swish function
class SwishImplementation(torch.autograd.Function):
    @staticmethod
    def forward(ctx, i):
        result = i * torch.sigmoid(i)
        ctx.save_for_backward(i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        i = ctx.saved_tensors[0]
        sigmoid_i = torch.sigmoid(i)
        return grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i)))

class MemoryEfficientSwish(nn.Module):
    def forward(self, x):
        return SwishImplementation.apply(x)


def round_filters(filters, global_params):
    """Calculate and round number of filters based on width multiplier.
       Use width_coefficient, depth_divisor and min_depth of global_params.

    Args:
        filters (int): Filters number to be calculated.
        global_params (namedtuple): Global params of the model.

    Returns:
        new_filters: New filters number after calculating.
    """
    multiplier = global_params.width_coefficient
    if not multiplier:
        return filters
    divisor = global_params.depth_divisor
    min_depth = global_params.min_depth
    filters *= multiplier
    min_depth = min_depth or divisor # pay attention to this line when using min_depth
    # follow the formula transferred from official TensorFlow implementation
    new_filters = max(min_depth, int(filters + divisor / 2) // divisor * divisor)
    if new_filters < 0.9 * filters: # prevent rounding by more than 10%
        new_filters += divisor
    return int(new_filters)


def round_repeats(repeats, global_params):
    """Calculate module's repeat number of a block based on depth multiplier.
       Use depth_coefficient of global_params.

    Args:
        repeats (int): num_repeat to be calculated.
        global_params (namedtuple): Global params of the model.

    Returns:
        new repeat: New repeat number after calculating.
    """
    multiplier = global_params.depth_coefficient
    if not multiplier:
        return repeats
    # follow the formula transferred from official TensorFlow implementation
    return int(math.ceil(multiplier * repeats))


def drop_connect(inputs, p, training):
    """Drop connect.

    Args:
        input (tensor: BCWH): Input of this structure.
        p (float: 0.0~1.0): Probability of drop connection.
        training (bool): The running mode.

    Returns:
        output: Output after drop connection.
    """
    assert 0 <= p <= 1, 'p must be in range of [0,1]'

    if not training:
        return inputs

    batch_size = inputs.shape[0]
    keep_prob = 1 - p

    # generate binary_tensor mask according to probability (p for 0, 1-p for 1)
    random_tensor = keep_prob
    random_tensor += torch.rand([batch_size, 1, 1, 1], dtype=inputs.dtype, device=inputs.device)
    binary_tensor = torch.floor(random_tensor)

    output = inputs / keep_prob * binary_tensor
    return output


def get_width_and_height_from_size(x):
    """Obtain height and width from x.

    Args:
        x (int, tuple or list): Data size.

    Returns:
        size: A tuple or list (H,W).
    """
    if isinstance(x, int):
        return x, x
    if isinstance(x, list) or isinstance(x, tuple):
        return x
    else:
        raise TypeError()


def calculate_output_image_size(input_image_size, stride):
    """Calculates the output image size when using Conv2dSamePadding with a stride.
       Necessary for static padding. Thanks to mannatsingh for pointing this out.

    Args:
        input_image_size (int, tuple or list): Size of input image.
        stride (int, tuple or list): Conv2d operation's stride.

    Returns:
        output_image_size: A list [H,W].
    """
    if input_image_size is None:
        return None
    image_height, image_width = get_width_and_height_from_size(input_image_size)
    stride = stride if isinstance(stride, int) else stride[0]
    image_height = int(math.ceil(image_height / stride))
    image_width = int(math.ceil(image_width / stride))
    return [image_height, image_width]


# Note:
# The following 'SamePadding' functions make output size equal ceil(input size/stride).
# Only when stride equals 1, can the output size be the same as input size.
# Don't be confused by their function names ! ! !

def get_same_padding_conv2d(image_size=None):
    """Chooses static padding if you have specified an image size, and dynamic padding otherwise.
       Static padding is necessary for ONNX exporting of models.

    Args:
        image_size (int or tuple): Size of the image.

    Returns:
        Conv2dDynamicSamePadding or Conv2dStaticSamePadding.
    """
    if image_size is None:
        return Conv2dDynamicSamePadding
    else:
        return partial(Conv2dStaticSamePadding, image_size=image_size)


class Conv2dDynamicSamePadding(nn.Conv2d):
    """2D Convolutions like TensorFlow, for a dynamic image size.
       The padding is operated in forward function by calculating dynamically.
    """

    # Tips for 'SAME' mode padding.
    #     Given the following:
    #         i: width or height
    #         s: stride
    #         k: kernel size
    #         d: dilation
    #         p: padding
    #     Output after Conv2d:
    #         o = floor((i+p-((k-1)*d+1))/s+1)
    # If o equals i, i = floor((i+p-((k-1)*d+1))/s+1),
    # => p = (i-1)*s+((k-1)*d+1)-i

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, dilation=1, groups=1, bias=True):
        super().__init__(in_channels, out_channels, kernel_size, stride, 0, dilation, groups, bias)
        self.stride = self.stride if len(self.stride) == 2 else [self.stride[0]] * 2

    def forward(self, x):
        ih, iw = x.size()[-2:]
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw) # change the output size according to stride ! ! !
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2])
        return F.conv2d(x, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


class Conv2dStaticSamePadding(nn.Conv2d):
    """2D Convolutions like TensorFlow's 'SAME' mode, with the given input image size.
       The padding module is calculated in construction function, then used in forward.
    """

    # With the same calculation as Conv2dDynamicSamePadding

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, image_size=None, **kwargs):
        super().__init__(in_channels, out_channels, kernel_size, stride, **kwargs)
        self.stride = self.stride if len(self.stride) == 2 else [self.stride[0]] * 2

        # Calculate padding based on image size and save it
        assert image_size is not None
        ih, iw = (image_size, image_size) if isinstance(image_size, int) else image_size
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            self.static_padding = nn.ZeroPad2d((pad_w - pad_w // 2, pad_w - pad_w // 2,
                                                pad_h - pad_h // 2, pad_h - pad_h // 2))
        else:
            self.static_padding = nn.Identity()

    def forward(self, x):
        x = self.static_padding(x)
        x = F.conv2d(x, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)
        return x


def get_same_padding_maxPool2d(image_size=None):
    """Chooses static padding if you have specified an image size, and dynamic padding otherwise.
       Static padding is necessary for ONNX exporting of models.

    Args:
        image_size (int or tuple): Size of the image.

    Returns:
        MaxPool2dDynamicSamePadding or MaxPool2dStaticSamePadding.
    """
    if image_size is None:
        return MaxPool2dDynamicSamePadding
    else:
        return partial(MaxPool2dStaticSamePadding, image_size=image_size)


class MaxPool2dDynamicSamePadding(nn.MaxPool2d):
    """2D MaxPooling like TensorFlow's 'SAME' mode, with a dynamic image size.
       The padding is operated in forward function by calculating dynamically.
    """

    def __init__(self, kernel_size, stride, padding=0, dilation=1, return_indices=False, ceil_mode=False):
        super().__init__(kernel_size, stride, padding, dilation, return_indices, ceil_mode)
        self.stride = [self.stride] * 2 if isinstance(self.stride, int) else self.stride
        self.kernel_size = [self.kernel_size] * 2 if isinstance(self.kernel_size, int) else self.kernel_size
        self.dilation = [self.dilation] * 2 if isinstance(self.dilation, int) else self.dilation

    def forward(self, x):
        ih, iw = x.size()[-2:]
        kh, kw = self.kernel_size
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2])
        return F.max_pool2d(x, self.kernel_size, self.stride, self.padding,
                            self.dilation, self.ceil_mode, self.return_indices)

class MaxPool2dStaticSamePadding(nn.MaxPool2d):
    """2D MaxPooling like TensorFlow's 'SAME' mode, with the given input image size.
       The padding module is calculated in construction function, then used in forward.
    """

    def __init__(self, kernel_size, stride, image_size=None, **kwargs):
        super().__init__(kernel_size, stride, **kwargs)
        self.stride = [self.stride] * 2 if isinstance(self.stride, int) else self.stride
        self.kernel_size = [self.kernel_size] * 2 if isinstance(self.kernel_size, int) else self.kernel_size
        self.dilation = [self.dilation] * 2 if isinstance(self.dilation, int) else self.dilation

        # Calculate padding based on image size and save it
        assert image_size is not None
        ih, iw = (image_size, image_size) if isinstance(image_size, int) else image_size
        kh, kw = self.kernel_size
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            self.static_padding = nn.ZeroPad2d((pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2))
        else:
            self.static_padding = nn.Identity()

    def forward(self, x):
        x = self.static_padding(x)
        x = F.max_pool2d(x, self.kernel_size, self.stride, self.padding,
                         self.dilation, self.ceil_mode, self.return_indices)
        return x


################################################################################
### Helper functions for loading model params
################################################################################

# BlockDecoder: A Class for encoding and decoding BlockArgs
# efficientnet_params: A function to query compound coefficient
# get_model_params and efficientnet:
#     Functions to get BlockArgs and GlobalParams for efficientnet
# url_map and url_map_advprop: Dicts of url_map for pretrained weights
# load_pretrained_weights: A function to load pretrained weights

class BlockDecoder(object):
    """Block Decoder for readability,
       straight from the official TensorFlow repository.
    """

    @staticmethod
    def _decode_block_string(block_string):
        """Get a block through a string notation of arguments.

        Args:
            block_string (str): A string notation of arguments.
                                Examples: 'r1_k3_s11_e1_i32_o16_se0.25_noskip'.

        Returns:
            BlockArgs: The namedtuple defined at the top of this file.
        """
        assert isinstance(block_string, str)

        ops = block_string.split('_')
        options = {}
        for op in ops:
            splits = re.split(r'(\d.*)', op)
            if len(splits) >= 2:
                key, value = splits[:2]
                options[key] = value

        # Check stride
        assert (('s' in options and len(options['s']) == 1) or
                (len(options['s']) == 2 and options['s'][0] == options['s'][1]))

        return BlockArgs(
            num_repeat=int(options['r']),
            kernel_size=int(options['k']),
            stride=[int(options['s'][0])],
            expand_ratio=int(options['e']),
            input_filters=int(options['i']),
            output_filters=int(options['o']),
            se_ratio=float(options['se']) if 'se' in options else None,
            id_skip=('noskip' not in block_string))

    @staticmethod
    def _encode_block_string(block):
        """Encode a block to a string.

        Args:
            block (namedtuple): A BlockArgs type argument.

        Returns:
            block_string: A String form of BlockArgs.
        """
        args = [
            'r%d' % block.num_repeat,
            'k%d' % block.kernel_size,
            's%d%d' % (block.strides[0], block.strides[1]),
            'e%s' % block.expand_ratio,
            'i%d' % block.input_filters,
            'o%d' % block.output_filters
        ]
        if 0 < block.se_ratio <= 1:
            args.append('se%s' % block.se_ratio)
        if block.id_skip is False:
            args.append('noskip')
        return '_'.join(args)

    @staticmethod
    def decode(string_list):
        """Decode a list of string notations to specify blocks inside the network.

        Args:
            string_list (list[str]): A list of strings, each string is a notation of block.

        Returns:
            blocks_args: A list of BlockArgs namedtuples of block args.
        """
        assert isinstance(string_list, list)
        blocks_args = []
        for block_string in string_list:
            blocks_args.append(BlockDecoder._decode_block_string(block_string))
        return blocks_args

    @staticmethod
    def encode(blocks_args):
        """Encode a list of BlockArgs to a list of strings.

        Args:
            blocks_args (list[namedtuples]): A list of BlockArgs namedtuples of block args.

        Returns:
            block_strings: A list of strings, each string is a notation of block.
        """
        block_strings = []
        for block in blocks_args:
            block_strings.append(BlockDecoder._encode_block_string(block))
        return block_strings


def efficientnet_params(model_name):
    """Map EfficientNet model name to parameter coefficients.

    Args:
        model_name (str): Model name to be queried.

    Returns:
        params_dict[model_name]: A (width,depth,res,dropout) tuple.
    """
    params_dict = {
        # Coefficients:   width,depth,res,dropout
        'efficientnet-b0': (1.0, 1.0, 224, 0.2),
        'efficientnet-b1': (1.0, 1.1, 240, 0.2),
        'efficientnet-b2': (1.1, 1.2, 260, 0.3),
        'efficientnet-b3': (1.2, 1.4, 300, 0.3),
        'efficientnet-b4': (1.4, 1.8, 380, 0.4),
        'efficientnet-b5': (1.6, 2.2, 456, 0.4),
        'efficientnet-b6': (1.8, 2.6, 528, 0.5),
        'efficientnet-b7': (2.0, 3.1, 600, 0.5),
        'efficientnet-b8': (2.2, 3.6, 672, 0.5),
        'efficientnet-l2': (4.3, 5.3, 800, 0.5),
    }
    return params_dict[model_name]


def efficientnet(width_coefficient=None, depth_coefficient=None, image_size=None,
                 dropout_rate=0.2, drop_connect_rate=0.2, num_classes=1000, include_top=True):
    """Create BlockArgs and GlobalParams for efficientnet model.

    Args:
        width_coefficient (float)
        depth_coefficient (float)
        image_size (int)
        dropout_rate (float)
        drop_connect_rate (float)
        num_classes (int)

        Meaning as the name suggests.

    Returns:
        blocks_args, global_params.
    """

    # Blocks args for the whole model(efficientnet-b0 by default)
    # It will be modified in the construction of EfficientNet Class according to model
    blocks_args = [
        'r1_k3_s11_e1_i32_o16_se0.25',
        'r2_k3_s22_e6_i16_o24_se0.25',
        'r2_k5_s22_e6_i24_o40_se0.25',
        'r3_k3_s22_e6_i40_o80_se0.25',
        'r3_k5_s11_e6_i80_o112_se0.25',
        'r4_k5_s22_e6_i112_o192_se0.25',
        'r1_k3_s11_e6_i192_o320_se0.25',
    ]
    blocks_args = BlockDecoder.decode(blocks_args)

    global_params = GlobalParams(
        width_coefficient=width_coefficient,
        depth_coefficient=depth_coefficient,
        image_size=image_size,
        dropout_rate=dropout_rate,

        num_classes=num_classes,
        batch_norm_momentum=0.99,
        batch_norm_epsilon=1e-3,
        drop_connect_rate=drop_connect_rate,
        depth_divisor=8,
        min_depth=None,
        include_top=include_top,
    )

    return blocks_args, global_params


def get_model_params(model_name, override_params):
    """Get the block args and global params for a given model name.

    Args:
        model_name (str): Model's name.
        override_params (dict): A dict to modify global_params.

    Returns:
        blocks_args, global_params
    """
    if model_name.startswith('efficientnet'):
        w, d, s, p = efficientnet_params(model_name)
        # note: all models have drop connect rate = 0.2
        blocks_args, global_params = efficientnet(
            width_coefficient=w, depth_coefficient=d, dropout_rate=p, image_size=s)
    else:
        raise NotImplementedError('model name is not pre-defined: {}'.format(model_name))
    if override_params:
        # ValueError will be raised here if override_params has fields not included in global_params.
        global_params = global_params._replace(**override_params)
    return blocks_args, global_params


# train with Standard methods
# check more details in paper(EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks)
url_map = {
    'efficientnet-b0': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b0-355c32eb.pth',
    'efficientnet-b1': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b1-f1951068.pth',
    'efficientnet-b2': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b2-8bb594d6.pth',
    'efficientnet-b3': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b3-5fb5a3c3.pth',
    'efficientnet-b4': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b4-6ed6700e.pth',
    'efficientnet-b5': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b5-b6417697.pth',
    'efficientnet-b6': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b6-c76e70fd.pth',
    'efficientnet-b7': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/efficientnet-b7-dcc49843.pth',
}

# train with Adversarial Examples(AdvProp)
# check more details in paper(Adversarial Examples Improve Image Recognition)
url_map_advprop = {
    'efficientnet-b0': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b0-b64d5a18.pth',
    'efficientnet-b1': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b1-0f3ce85a.pth',
    'efficientnet-b2': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b2-6e9d97e5.pth',
    'efficientnet-b3': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b3-cdd7c0f4.pth',
    'efficientnet-b4': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b4-44fb3a87.pth',
    'efficientnet-b5': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b5-86493f6b.pth',
    'efficientnet-b6': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b6-ac80338e.pth',
    'efficientnet-b7': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b7-4652b6dd.pth',
    'efficientnet-b8': 'https://github.com/lukemelas/EfficientNet-PyTorch/releases/download/1.0/adv-efficientnet-b8-22a8fe65.pth',
}

# TODO: add the petrained weights url map of 'efficientnet-l2'


def load_pretrained_weights(model, model_name, weights_path=None, load_fc=True, advprop=False):
    """Loads pretrained weights from weights path or download using url.

    Args:
        model (Module): The whole model of efficientnet.
        model_name (str): Model name of efficientnet.
        weights_path (None or str):
            str: path to pretrained weights file on the local disk.
            None: use pretrained weights downloaded from the Internet.
        load_fc (bool): Whether to load pretrained weights for fc layer at the end of the model.
        advprop (bool): Whether to load pretrained weights
                        trained with advprop (valid when weights_path is None).
    """
    if isinstance(weights_path, str):
        state_dict = torch.load(weights_path)
    else:
        # AutoAugment or Advprop (different preprocessing)
        url_map_ = url_map_advprop if advprop else url_map
        state_dict = model_zoo.load_url(url_map_[model_name])

    if load_fc:
        ret = model.load_state_dict(state_dict, strict=False)
        assert not ret.missing_keys, 'Missing keys when loading pretrained weights: {}'.format(ret.missing_keys)
    else:
        state_dict.pop('_fc.weight')
        state_dict.pop('_fc.bias')
        ret = model.load_state_dict(state_dict, strict=False)
        assert set(ret.missing_keys) == set(
            ['_fc.weight', '_fc.bias']), 'Missing keys when loading pretrained weights: {}'.format(ret.missing_keys)
    assert not ret.unexpected_keys, 'Missing keys when loading pretrained weights: {}'.format(ret.unexpected_keys)

    print('Loaded pretrained weights for {}'.format(model_name))



"""

render_images_with_thumbnails.py

Renders an output image with one primary image and crops from many secondary images,
used primarily to check whether candidate repeat detections are actually false positives or not.

"""

#%% Imports

import math
import os
import random

from PIL import Image

from megadetector.visualization import visualization_utils as vis_utils
from megadetector.utils import path_utils


#%% Support functions

def crop_image_with_normalized_coordinates(
        image,
        bounding_box):
    """
    Args:
        image (PIL.Image): image to crop
        bounding_box (tuple): tuple formatted as (x,y,w,h), where (0,0) is the
            upper-left of the image, and coordinates are normalized
            (so (0,0,1,1) is a box containing the entire image).
            
    Returns:
        PIL.Image: cropped image
    """
    
    im_width, im_height = image.size
    (x_norm, y_norm, w_norm, h_norm) = bounding_box
    (x, y, w, h) = (x_norm * im_width,
                    y_norm * im_height,
                    w_norm * im_width,
                    h_norm * im_height)
    return image.crop((x, y, x+w, y+h))


#%% Main function

def render_images_with_thumbnails(
        primary_image_filename,
        primary_image_width,
        secondary_image_filename_list,
        secondary_image_bounding_box_list,
        cropped_grid_width,
        output_image_filename,
        primary_image_location='right'):
    """
    Given a primary image filename and a list of secondary images, writes to
    the provided output_image_filename an image where the one
    side is the primary image, and the other side is a grid of the 
    secondary images, cropped according to the provided list of bounding
    boxes.

    The output file will be primary_image_width + cropped_grid_width pixels
    wide.

    The height of the output image will be determined by the original aspect
    ratio of the primary image. 
    
    Args:
        primary_image_filename (str): filename of the primary image to load as str
        primary_image_width (int): width at which to render the primary image; if this is 
            None, will render at the original image width
        secondary_image_filename_list (list): list of filenames of the secondary images
        secondary_image_bounding_box_list (list): list of tuples, one per secondary
            image. Each tuple is a bounding box of the secondary image,
            formatted as (x,y,w,h), where (0,0) is the upper-left of the image,
            and coordinates are normalized (so (0,0,1,1) is a box containing
            the entire image.
        cropped_grid_width (int): width of the cropped-image area
        output_image_filename (str): filename to write the output image        
        primary_image_location (str, optional): 'right' or left'; reserving 'top', 'bottom', etc.
            for future use
    """

    # Check to make sure the arguments are reasonable
    assert(len(secondary_image_filename_list) ==
           len(secondary_image_bounding_box_list)), \
           'Length of secondary image list and bounding box list should be equal'

    assert primary_image_location in ['left','right']
    
    # Load primary image and resize to desired width
    primary_image = vis_utils.load_image(primary_image_filename)
    if primary_image_width is not None:
        primary_image = vis_utils.resize_image(primary_image, primary_image_width, 
                                               target_height=-1)

    # Compute the number of grid elements for the secondary images
    # to best fit the available aspect ratio
    grid_width = cropped_grid_width
    grid_height = primary_image.size[1]
    grid_aspect = grid_width / grid_height
    
    sample_crop_width = secondary_image_bounding_box_list[0][2]
    sample_crop_height = secondary_image_bounding_box_list[0][3]
    
    n_crops = len(secondary_image_filename_list)
    
    optimal_n_rows = None
    optimal_aspect_error = None
    
    for candidate_n_rows in range(1,n_crops+1):
        candidate_n_cols = math.ceil(n_crops / candidate_n_rows)
        candidate_grid_aspect = (candidate_n_cols*sample_crop_width) / \
          (candidate_n_rows*sample_crop_height)
        aspect_error = abs(grid_aspect-candidate_grid_aspect)
        if optimal_n_rows is None or aspect_error < optimal_aspect_error:
            optimal_n_rows = candidate_n_rows
            optimal_aspect_error = aspect_error
    
    assert optimal_n_rows is not None
    grid_rows = optimal_n_rows
    grid_columns = math.ceil(n_crops/grid_rows)        
    
    # Compute the width of each grid cell
    grid_cell_width = math.floor(grid_width / grid_columns)
    grid_cell_height = math.floor(grid_height / grid_rows)
    
    # Load secondary images and their associated bounding boxes. Iterate
    # through them, crop them, and save them to a list of cropped_images
    cropped_images = []
    for (name, box) in zip(secondary_image_filename_list,
                           secondary_image_bounding_box_list):
        
        other_image = vis_utils.load_image(name)
        cropped_image = crop_image_with_normalized_coordinates(
                other_image, box)
        
        # Rescale this crop to fit within the desired grid cell size
        width_scale_factor = grid_cell_width / cropped_image.size[0]
        height_scale_factor = grid_cell_height / cropped_image.size[1]
        scale_factor = min(width_scale_factor,height_scale_factor)
        
        # Resize the cropped image, whether we're making it larger or smaller
        cropped_image = cropped_image.resize(
                ((int)(cropped_image.size[0] * scale_factor),
                 (int)(cropped_image.size[1] * scale_factor)))            

        cropped_images.append(cropped_image)        

    # ...for each crop
    
    # Compute the final output image size. This will depend upon the aspect
    # ratio of the crops.
    output_image_width = primary_image.size[0] + grid_width
    output_image_height = primary_image.size[1]

    # Create blank output image
    output_image = Image.new('RGB', (output_image_width, output_image_height))

    # Copy resized primary image to output image
    if primary_image_location == 'right':
        primary_image_x = grid_width
    else:
        primary_image_x = 0
        
    output_image.paste(primary_image, (primary_image_x, 0))

    # Compute the final locations of the secondary images in the output image
    i_row = 0; i_col = 0
    for image in cropped_images:
        
        x = i_col * grid_cell_width
        if primary_image_location == 'left':
            x += primary_image.size[0]
        y = i_row * grid_cell_height 
        output_image.paste(image, (x,y))
        i_col += 1
        if i_col >= grid_columns:
            i_col = 0
            i_row += 1
            
    # ...for each crop

    # Write output image to disk
    output_image.save(output_image_filename)    

# ...def render_images_with_thumbnails(...)


#%% Interactive driver

if False:
    
    pass

    #%%
    
    primary_image_filename = '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0035.JPG'
    
    primary_image_width = 5152
    
    secondary_image_filename_list = ['/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0035.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0040.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0007.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0041.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0008.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0048.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0031.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0006.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0004.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0026.JPG', '/home/user/data/KRU/KRU_public/KRU_S1/13/13_R1/KRU_S1_13_R1_IMAG0005.JPG']
    
    secondary_image_bounding_box_list = [[0, 0, 0.1853, 0.6552], [0, 0, 0.1855, 0.6527], [0, 0.000252, 0.1991, 0.6925], [0, 0, 0.1855, 0.6527], [0, 0.001008, 0.1902, 0.6774], [0, 0, 0.1845, 0.658], [0, 0, 0.1824, 0.6711], [0, 0.00252, 0.2005, 0.6857], [0, 0.002268, 0.1983, 0.6852], [0, 0, 0.1752, 0.6897], [0, 0.001764, 0.1989, 0.6887]]
    
    # cropped_grid_width = 3091
    cropped_grid_width = 500
    
    primary_image_location = 'right'

    output_image_filename = os.path.expanduser('~/tmp/grid-test.jpg')
    
    render_images_with_thumbnails(
            primary_image_filename,
            primary_image_width,
            secondary_image_filename_list,
            secondary_image_bounding_box_list,
            cropped_grid_width,
            output_image_filename,
            primary_image_location='right')    

    path_utils.open_file(output_image_filename)
    
    
#%% Command-line driver

# This is just a test driver, this module is not meant to be run from the command line.

def main():
        
    # Load images from a test directory.
    #
    # Make the first image in the directory the primary image, 
    # the remaining ones the comparison images.    
    test_input_folder = os.path.expanduser('~/data/KRU-test')
    output_image_filename = os.path.expanduser('~/tmp/thumbnail_test.jpg')
    
    files = path_utils.find_images(test_input_folder)    

    random.seed(0); random.shuffle(files)
    primary_image_filename = files[0]

    secondary_image_filename_list = []
    secondary_image_bounding_box_list = []

    # Initialize the x,y location of the bounding box
    box = (random.uniform(0.25, 0.75), random.uniform(0.25, 0.75))

    # Create the list of secondary images and their bounding boxes
    for file in files[1:]:
        secondary_image_filename_list.append(file)
        secondary_image_bounding_box_list.append(
                (box[0] + random.uniform(-0.001, 0.001),
                 box[1] + random.uniform(-0.001, 0.001),
                 0.2,
                 0.2))
    
    primary_image_width = 1000
    cropped_grid_width = 1000

    render_images_with_thumbnails(
        primary_image_filename,
        primary_image_width,
        secondary_image_filename_list,
        secondary_image_bounding_box_list,
        cropped_grid_width,
        output_image_filename, 'right')
    
    from megadetector.utils import path_utils
    path_utils.open_file(output_image_filename)
        
if __name__ == '__main__':
    main()



"""

generate_lila_per_image_labels.py

Generate a .csv file with one row per annotation, containing full URLs to every
camera trap image on LILA, with taxonomically expanded labels.

Typically there will be one row per image, though images with multiple annotations
will have multiple rows.

Some images may not physically exist, particularly images that are labeled as "human".
This script does not validate image URLs.

Does not include bounding box annotations.

"""

#%% Constants and imports

import os
import json
import pandas as pd
import numpy as np
import dateparser
import csv

from collections import defaultdict
from tqdm import tqdm

from megadetector.data_management.lila.lila_common import \
    read_lila_metadata, \
    read_metadata_file_for_dataset, \
    read_lila_taxonomy_mapping

from megadetector.utils import write_html_image_list
from megadetector.utils.path_utils import zip_file
from megadetector.utils.path_utils import open_file

# We'll write images, metadata downloads, and temporary files here
lila_local_base = os.path.expanduser('~/lila')
preview_folder = os.path.join(lila_local_base,'csv_preview')

os.makedirs(lila_local_base,exist_ok=True)

metadata_dir = os.path.join(lila_local_base,'metadata')
os.makedirs(metadata_dir,exist_ok=True)

output_file = os.path.join(lila_local_base,'lila_image_urls_and_labels.csv')

# Some datasets don't have "sequence_level_annotation" fields populated, but we know their 
# annotation level
ds_name_to_annotation_level = {}
ds_name_to_annotation_level['Caltech Camera Traps'] = 'image'
ds_name_to_annotation_level['ENA24'] = 'image'
ds_name_to_annotation_level['Island Conservation Camera Traps'] = 'image'
ds_name_to_annotation_level['Channel IslandsCamera Traps'] = 'image'
ds_name_to_annotation_level['WCS Camera Traps'] = 'sequence'
ds_name_to_annotation_level['Wellington Camera Traps'] = 'sequence'
ds_name_to_annotation_level['NACTI'] = 'unknown'

known_unmapped_labels = set(['WCS Camera Traps:#ref!'])

debug_max_images_per_dataset = -1
if debug_max_images_per_dataset > 0:
    print('Running in debug mode')
    output_file = output_file.replace('.csv','_debug.csv')


#%% Download and parse the metadata file

metadata_table = read_lila_metadata(metadata_dir)

# To select an individual data set for debugging
if False:
    k = 'Idaho Camera Traps'
    metadata_table = {k:metadata_table[k]}


#%% Download and extract metadata for each dataset

for ds_name in metadata_table.keys():    
    metadata_table[ds_name]['metadata_filename'] = read_metadata_file_for_dataset(ds_name=ds_name,
                                                                         metadata_dir=metadata_dir,
                                                                         metadata_table=metadata_table)
    
#%% Load taxonomy data

taxonomy_df = read_lila_taxonomy_mapping(metadata_dir)


#%% Build a dictionary that maps each [dataset,query] pair to the full taxonomic label set

ds_label_to_taxonomy = {}

# i_row = 0; row = taxonomy_df.iloc[i_row]
for i_row,row in taxonomy_df.iterrows():
    
    ds_label = row['dataset_name'] + ':' + row['query']
    assert ds_label.strip() == ds_label
    assert ds_label not in ds_label_to_taxonomy
    ds_label_to_taxonomy[ds_label] = row.to_dict()
    

#%% Process annotations for each dataset

# Takes several hours

# The order of these headers needs to match the order in which fields are added later in this cell;
# don't mess with this order.
header = ['dataset_name','url_gcp','url_aws','url_azure',
          'image_id','sequence_id','location_id','frame_num',
          'original_label','scientific_name','common_name','datetime','annotation_level']

taxonomy_levels_to_include = \
    ['kingdom','phylum','subphylum','superclass','class','subclass','infraclass','superorder','order',
     'suborder','infraorder','superfamily','family','subfamily','tribe','genus','species','subspecies',
     'variety']
    
header.extend(taxonomy_levels_to_include)

missing_annotations = set()

def clearnan(v):
    if isinstance(v,float):
        assert np.isnan(v)
        v = ''
    assert isinstance(v,str)
    return v

with open(output_file,'w',encoding='utf-8',newline='') as f:
    
    csv_writer = csv.writer(f)
    csv_writer.writerow(header)
    
    # ds_name = list(metadata_table.keys())[0]
    for ds_name in metadata_table.keys():
        
        if 'bbox' in ds_name:
            print('Skipping bbox dataset {}'.format(ds_name))
            continue
    
        print('Processing dataset {}'.format(ds_name))
        
        json_filename = metadata_table[ds_name]['metadata_filename']
        with open(json_filename, 'r') as f:
            data = json.load(f)
        
        categories = data['categories']
        category_ids = [c['id'] for c in categories]
        for c in categories:
            category_id_to_name = {c['id']:c['name'] for c in categories}
            
        annotations = data['annotations']
        images = data['images']
        
        image_id_to_annotations = defaultdict(list)
        
        # Go through annotations, marking each image with the categories that are present
        #
        # ann = annotations[0]
        for ann in annotations:            
            image_id_to_annotations[ann['image_id']].append(ann)
    
        unannotated_images = []
        
        found_date = False
        found_location = False
        found_annotation_level = False
        
        if ds_name in ds_name_to_annotation_level:
            expected_annotation_level = ds_name_to_annotation_level[ds_name]
        else:
            expected_annotation_level = None
                    
        # im = images[10]
        for i_image,im in enumerate(images):
            
            if (debug_max_images_per_dataset is not None) and (debug_max_images_per_dataset > 0) \
                and (i_image >= debug_max_images_per_dataset):
                break
            
            file_name = im['file_name'].replace('\\','/')
            base_url_gcp = metadata_table[ds_name]['image_base_url_gcp']
            base_url_aws = metadata_table[ds_name]['image_base_url_aws']
            base_url_azure = metadata_table[ds_name]['image_base_url_azure']
            assert not base_url_gcp.endswith('/')
            assert not base_url_aws.endswith('/')
            assert not base_url_azure.endswith('/')
            
            url_gcp = base_url_gcp + '/' + file_name
            url_aws = base_url_aws + '/' + file_name
            url_azure = base_url_azure + '/' + file_name
                        
            for k in im.keys():
                if ('date' in k or 'time' in k) and (k not in ['datetime','date_captured']):
                    raise ValueError('Unrecognized datetime field')
                    
            # This field name was only used for Caltech Camera Traps
            if 'date_captured' in im:
                assert ds_name == 'Caltech Camera Traps'
                im['datetime'] = im['date_captured']
                
            def has_valid_datetime(im):
                if 'datetime' not in im:
                    return False
                v = im['datetime']
                if v is None:
                    return False
                if isinstance(v,str):
                    return len(v) > 0
                else:
                    assert isinstance(v,float) and np.isnan(v)
                    return False
                    
            dt_string = ''                
            if (has_valid_datetime(im)):
                
                dt = dateparser.parse(im['datetime'])
                
                if dt is None or dt.year < 1990 or dt.year > 2025:
                    
                    # raise ValueError('Suspicious date parsing result')
                    
                    # Special case we don't want to print a warning about... this is 
                    # in invalid date that very likely originates on the camera, not at
                    # some intermediate processing step.
                    #
                    # print('Suspicious date for image {}: {} ({})'.format(
                    #    im['id'], im['datetime'], ds_name))
                    pass                
                    
                else:
                    
                    found_date = True
                    dt_string = dt.strftime("%m-%d-%Y %H:%M:%S")
                
            # Location, sequence, and image IDs are only guaranteed to be unique within
            # a dataset, so for the output .csv file, include both
            if 'location' in im:
                found_location = True
                location_id = ds_name + ' : ' + str(im['location'])
            else:
                location_id = ds_name
                
            image_id = ds_name + ' : ' + str(im['id'])
            
            if 'seq_id' in im:
                sequence_id = ds_name + ' : ' + str(im['seq_id'])
            else:
                sequence_id = ds_name + ' : ' + 'unknown'
                
            if 'frame_num' in im:
                frame_num = im['frame_num']
            else:
                frame_num = -1
            
            annotations_this_image = image_id_to_annotations[im['id']]
            
            categories_this_image = set()
            
            annotation_level = 'unknown'
            
            for ann in annotations_this_image:
                assert ann['image_id'] == im['id']
                categories_this_image.add(category_id_to_name[ann['category_id']])
                if 'sequence_level_annotation' in ann:
                    found_annotation_level = True
                    if ann['sequence_level_annotation']:
                        annotation_level = 'sequence'
                    else:
                        annotation_level = 'image'
                    if expected_annotation_level is not None:
                        assert expected_annotation_level == annotation_level,\
                            'Unexpected annotation level'
                elif expected_annotation_level is not None:
                    annotation_level = expected_annotation_level
                    
            if len(categories_this_image) == 0:
                unannotated_images.append(im)
                continue
            
            # category_name = list(categories_this_image)[0]
            for category_name in categories_this_image:
                
                ds_label = ds_name + ':' + category_name.lower()
                
                if ds_label not in ds_label_to_taxonomy:
                    
                    assert ds_label in known_unmapped_labels
                    
                    # Only print a warning the first time we see an unmapped label
                    if ds_label not in missing_annotations:
                        print('Warning: {} not in taxonomy file'.format(ds_label))
                    missing_annotations.add(ds_label)
                    continue
                
                taxonomy_labels = ds_label_to_taxonomy[ds_label]
                
                """
                header =        
                    ['dataset_name','url','image_id','sequence_id','location_id',
                     'frame_num','original_label','scientific_name','common_name',
                     'datetime','annotation_level']
                """
                
                row = []
                row.append(ds_name)
                row.append(url_gcp)
                row.append(url_aws)
                row.append(url_azure)
                row.append(image_id)
                row.append(sequence_id)
                row.append(location_id)
                row.append(frame_num)
                row.append(taxonomy_labels['query'])
                row.append(clearnan(taxonomy_labels['scientific_name']))
                row.append(clearnan(taxonomy_labels['common_name']))
                row.append(dt_string)
                row.append(annotation_level)
                
                for s in taxonomy_levels_to_include:
                    row.append(clearnan(taxonomy_labels[s]))
                
                assert len(row) == len(header)
                
                csv_writer.writerow(row)
                        
            # ...for each category that was applied at least once to this image
            
        # ...for each image in this dataset
        
        if not found_date:
            pass
            # print('Warning: no date information available for this dataset')
            
        if not found_location:
            pass
            # print('Warning: no location information available for this dataset')
        
        if not found_annotation_level and (ds_name not in ds_name_to_annotation_level):
            print('Warning: no annotation level information available for this dataset')
        
        if len(unannotated_images) > 0:
            print('Warning: {} of {} images are un-annotated\n'.\
                  format(len(unannotated_images),len(images)))
        
    # ...for each dataset

# ...with open()

print('Processed {} datasets'.format(len(metadata_table)))


#%% Read the .csv back

df = pd.read_csv(output_file)
print('Read {} lines from {}'.format(len(df),output_file))


#%% Do some post-hoc integrity checking

# Takes ~10 minutes without using apply()

tqdm.pandas()

def isint(v):
    return isinstance(v,int) or isinstance(v,np.int64)

valid_annotation_levels = set(['sequence','image','unknown'])

# Collect a list of locations within each dataset; we'll use this
# in the next cell to look for datasets that only have a single location
dataset_name_to_locations = defaultdict(set)

def check_row(row):
    
    assert row['dataset_name'] in metadata_table.keys()
    for url_column in ['url_gcp','url_aws','url_azure']:
        assert row[url_column].startswith('https://') or row[url_column].startswith('http://')
    assert ' : ' in row['image_id']
    assert 'seq' not in row['location_id'].lower()
    assert row['annotation_level'] in valid_annotation_levels

    # frame_num should either be NaN or an integer
    if isinstance(row['frame_num'],float):
        assert np.isnan(row['frame_num'])
    else:
        # -1 is sometimes used for sequences of unknown length
        assert isint(row['frame_num']) and row['frame_num'] >= -1

    ds_name = row['dataset_name']
    dataset_name_to_locations[ds_name].add(row['location_id'])
    
# Faster, but more annoying to debug
if False:
    
    df.progress_apply(check_row, axis=1)

else:
    
    # i_row = 0; row = df.iloc[i_row]
    for i_row,row in tqdm(df.iterrows(),total=len(df)):
        check_row(row)


#%% Check for datasets that have only one location string

# Expected: ENA24, Missouri Camera Traps

for ds_name in dataset_name_to_locations.keys():
    if len(dataset_name_to_locations[ds_name]) == 1:
        print('No location information for {}'.format(ds_name))


#%% Preview constants

n_empty_images_per_dataset = 3
n_non_empty_images_per_dataset = 10

os.makedirs(preview_folder,exist_ok=True)


#%% Choose images to download

np.random.seed(0)
images_to_download = []

# ds_name = list(metadata_table.keys())[2]
for ds_name in metadata_table.keys():
    
    if 'bbox' in ds_name:
        continue
    
    # Find all rows for this dataset
    ds_rows = df.loc[df['dataset_name'] == ds_name]
    
    print('{} rows available for {}'.format(len(ds_rows),ds_name))
    assert len(ds_rows) > 0
    
    empty_rows = ds_rows[ds_rows['scientific_name'].isnull()]
    non_empty_rows = ds_rows[~ds_rows['scientific_name'].isnull()]
    
    if len(empty_rows) == 0:
        print('No empty images available for {}'.format(ds_name))
    elif len(empty_rows) > n_empty_images_per_dataset:
        empty_rows = empty_rows.sample(n=n_empty_images_per_dataset)
    images_to_download.extend(empty_rows.to_dict('records'))

    if len(non_empty_rows) == 0:
        print('No non-empty images available for {}'.format(ds_name))
    elif len(non_empty_rows) > n_non_empty_images_per_dataset:
        non_empty_rows = non_empty_rows.sample(n=n_non_empty_images_per_dataset)
    images_to_download.extend(non_empty_rows.to_dict('records'))
    
 # ...for each dataset

print('Selected {} total images'.format(len(images_to_download)))


#%% Download images (prep)

# Expect a few errors for images with human or vehicle labels (or things like "ignore" that *could* be humans)

preferred_cloud = 'aws'

url_to_target_file = {}

# i_image = 10; image = images_to_download[i_image]
for i_image,image in tqdm(enumerate(images_to_download),total=len(images_to_download)):
    
    url = image['url_' + preferred_cloud]
    ext = os.path.splitext(url)[1]
    fn_relative = 'image_{}'.format(str(i_image).zfill(4)) + ext
    fn_abs = os.path.join(preview_folder,fn_relative)
    image['relative_file'] = fn_relative
    image['url'] = url
    url_to_target_file[url] = fn_abs
    

#%% Download images (execution)

from megadetector.utils.url_utils import parallel_download_urls
download_results = parallel_download_urls(url_to_target_file,verbose=False,overwrite=True,
                                          n_workers=20,pool_type='thread')


#%% Write preview HTML

html_filename = os.path.join(preview_folder,'index.html')

html_images = []

# im = images_to_download[0]
for im in images_to_download:
    
    if im['relative_file'] is None:
        continue
    
    output_im = {}
    output_im['filename'] = im['relative_file']
    output_im['linkTarget'] = im['url']
    output_im['title'] = '<b>{}: {}</b><br/><br/>'.format(im['dataset_name'],im['original_label']) + str(im)
    output_im['imageStyle'] = 'width:600px;'
    output_im['textStyle'] = 'font-weight:normal;font-size:100%;'
    html_images.append(output_im)
    
write_html_image_list.write_html_image_list(html_filename,html_images)

open_file(html_filename)


#%% Zip output file

zipped_output_file = zip_file(output_file,verbose=True)

print('Zipped {} to {}'.format(output_file,zipped_output_file))



import io
import json
import math
import os
import sys
from datetime import datetime
from io import BytesIO
from typing import Union

from PIL import Image
import numpy as np
import requests
import tensorflow as tf
from azure.storage.blob import ContainerClient

print('score.py, tensorflow version:', tf.__version__)
print('score.py, tf.test.is_gpu_available:', tf.test.is_gpu_available())

PRINT_EVERY = 500


#%% Helper functions *copied* from ct_utils.py and visualization/visualization_utils.py

IMAGE_ROTATIONS = {
    3: 180,
    6: 270,
    8: 90
}

def truncate_float(x, precision=3):
    """
    Function for truncating a float scalar to the defined precision.
    For example: truncate_float(0.0003214884) --> 0.000321
    This function is primarily used to achieve a certain float representation
    before exporting to JSON
    Args:
    x         (float) Scalar to truncate
    precision (int)   The number of significant digits to preserve, should be
                      greater or equal 1
    """

    assert precision > 0

    if np.isclose(x, 0):
        return 0
    else:
        # Determine the factor, which shifts the decimal point of x
        # just behind the last significant digit
        factor = math.pow(10, precision - 1 - math.floor(math.log10(abs(x))))
        # Shift decimal point by multiplicatipon with factor, flooring, and
        # division by factor
        return math.floor(x * factor)/factor


def open_image(input_file: Union[str, BytesIO]) -> Image:
    """Opens an image in binary format using PIL.Image and converts to RGB mode.

    This operation is lazy; image will not be actually loaded until the first
    operation that needs to load it (for example, resizing), so file opening
    errors can show up later.

    Args:
        input_file: str or BytesIO, either a path to an image file (anything
            that PIL can open), or an image as a stream of bytes

    Returns:
        an PIL image object in RGB mode
    """
    if (isinstance(input_file, str)
            and input_file.startswith(('http://', 'https://'))):
        response = requests.get(input_file)
        image = Image.open(BytesIO(response.content))
        try:
            response = requests.get(input_file)
            image = Image.open(BytesIO(response.content))
        except Exception as e:
            print(f'Error opening image {input_file}: {e}')
            raise
    else:
        image = Image.open(input_file)
    if image.mode not in ('RGBA', 'RGB', 'L'):
        raise AttributeError(f'Image {input_file} uses unsupported mode {image.mode}')
    if image.mode == 'RGBA' or image.mode == 'L':
        # PIL.Image.convert() returns a converted copy of this image
        image = image.convert(mode='RGB')

    # alter orientation as needed according to EXIF tag 0x112 (274) for Orientation
    # https://gist.github.com/dangtrinhnt/a577ece4cbe5364aad28
    # https://www.media.mit.edu/pia/Research/deepview/exif.html
    try:
        exif = image._getexif()
        orientation: int = exif.get(274, None)  # 274 is the key for the Orientation field
        if orientation is not None and orientation in IMAGE_ROTATIONS:
            image = image.rotate(IMAGE_ROTATIONS[orientation], expand=True)  # returns a rotated copy
    except Exception:
        pass

    return image


def load_image(input_file: Union[str, BytesIO]) -> Image.Image:
    """Loads the image at input_file as a PIL Image into memory.
    Image.open() used in open_image() is lazy and errors will occur downstream
    if not explicitly loaded.
    Args:
        input_file: str or BytesIO, either a path to an image file (anything
            that PIL can open), or an image as a stream of bytes
    Returns: PIL.Image.Image, in RGB mode
    """
    image = open_image(input_file)
    image.load()
    return image


#%% TFDetector class, an unmodified *copy* of the class in detection/tf_detector.py,
# so we do not have to import the packages required by run_detector.py

class TFDetector:
    """
    A detector model loaded at the time of initialization. It is intended to be used with
    MegaDetector (TF). The inference batch size is set to 1; code needs to be modified
    to support larger batch sizes, including resizing appropriately.
    """

    # Number of decimal places to round to for confidence and bbox coordinates
    CONF_DIGITS = 3
    COORD_DIGITS = 4

    # MegaDetector was trained with batch size of 1, and the resizing function is a part
    # of the inference graph
    BATCH_SIZE = 1

    # An enumeration of failure reasons
    FAILURE_TF_INFER = 'Failure TF inference'
    FAILURE_IMAGE_OPEN = 'Failure image access'

    DEFAULT_RENDERING_CONFIDENCE_THRESHOLD = 0.85  # to render bounding boxes
    DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD = 0.1  # to include in the output json file

    DEFAULT_DETECTOR_LABEL_MAP = {
        '1': 'animal',
        '2': 'person',
        '3': 'vehicle'  # available in megadetector v4+
    }

    NUM_DETECTOR_CATEGORIES = 4  # animal, person, group, vehicle - for color assignment

    def __init__(self, model_path):
        """Loads model from model_path and starts a tf.Session with this graph. Obtains
        input and output tensor handles."""
        detection_graph = TFDetector.__load_model(model_path)
        self.tf_session = tf.Session(graph=detection_graph)

        self.image_tensor = detection_graph.get_tensor_by_name('image_tensor:0')
        self.box_tensor = detection_graph.get_tensor_by_name('detection_boxes:0')
        self.score_tensor = detection_graph.get_tensor_by_name('detection_scores:0')
        self.class_tensor = detection_graph.get_tensor_by_name('detection_classes:0')

    @staticmethod
    def round_and_make_float(d, precision=4):
        return truncate_float(float(d), precision=precision)

    @staticmethod
    def __convert_coords(tf_coords):
        """Converts coordinates from the model's output format [y1, x1, y2, x2] to the
        format used by our API and MegaDB: [x1, y1, width, height]. All coordinates
        (including model outputs) are normalized in the range [0, 1].
        Args:
            tf_coords: np.array of predicted bounding box coordinates from the TF detector,
                has format [y1, x1, y2, x2]
        Returns: list of Python float, predicted bounding box coordinates [x1, y1, width, height]
        """
        # change from [y1, x1, y2, x2] to [x1, y1, width, height]
        width = tf_coords[3] - tf_coords[1]
        height = tf_coords[2] - tf_coords[0]

        new = [tf_coords[1], tf_coords[0], width, height]  # must be a list instead of np.array

        # convert numpy floats to Python floats
        for i, d in enumerate(new):
            new[i] = TFDetector.round_and_make_float(d, precision=TFDetector.COORD_DIGITS)
        return new

    @staticmethod
    def convert_to_tf_coords(array):
        """From [x1, y1, width, height] to [y1, x1, y2, x2], where x1 is x_min, x2 is x_max
        This is an extraneous step as the model outputs [y1, x1, y2, x2] but were converted to the API
        output format - only to keep the interface of the sync API.
        """
        x1 = array[0]
        y1 = array[1]
        width = array[2]
        height = array[3]
        x2 = x1 + width
        y2 = y1 + height
        return [y1, x1, y2, x2]

    @staticmethod
    def __load_model(model_path):
        """Loads a detection model (i.e., create a graph) from a .pb file.
        Args:
            model_path: .pb file of the model.
        Returns: the loaded graph.
        """
        print('TFDetector: Loading graph...')
        detection_graph = tf.Graph()
        with detection_graph.as_default():
            od_graph_def = tf.GraphDef()
            with tf.gfile.GFile(model_path, 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')
        print('TFDetector: Detection graph loaded.')

        return detection_graph

    def _generate_detections_one_image(self, image):
        np_im = np.asarray(image, np.uint8)
        im_w_batch_dim = np.expand_dims(np_im, axis=0)

        # need to change the above line to the following if supporting a batch size > 1 and resizing to the same size
        # np_images = [np.asarray(image, np.uint8) for image in images]
        # images_stacked = np.stack(np_images, axis=0) if len(images) > 1 else np.expand_dims(np_images[0], axis=0)

        # performs inference
        (box_tensor_out, score_tensor_out, class_tensor_out) = self.tf_session.run(
            [self.box_tensor, self.score_tensor, self.class_tensor],
            feed_dict={self.image_tensor: im_w_batch_dim})

        return box_tensor_out, score_tensor_out, class_tensor_out

    def generate_detections_one_image(self, image, image_id,
                                      detection_threshold=DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD):
        """Apply the detector to an image.
        Args:
            image: the PIL Image object
            image_id: a path to identify the image; will be in the "file" field of the output object
            detection_threshold: confidence above which to include the detection proposal
        Returns:
        A dict with the following fields, see the 'images' key in https://github.com/agentmorris/MegaDetector/tree/main/megadetector/api/batch_processing#batch-processing-api-output-format
            - 'file' (always present)
            - 'max_detection_conf'
            - 'detections', which is a list of detection objects containing keys 'category', 'conf' and 'bbox'
            - 'failure'
        """
        result = {
            'file': image_id
        }
        try:
            b_box, b_score, b_class = self._generate_detections_one_image(image)

            # our batch size is 1; need to loop the batch dim if supporting batch size > 1
            boxes, scores, classes = b_box[0], b_score[0], b_class[0]

            detections_cur_image = []  # will be empty for an image with no confident detections
            max_detection_conf = 0.0
            for b, s, c in zip(boxes, scores, classes):
                if s > detection_threshold:
                    detection_entry = {
                        'category': str(int(c)),  # use string type for the numerical class label, not int
                        'conf': truncate_float(float(s),  # cast to float for json serialization
                                               precision=TFDetector.CONF_DIGITS),
                        'bbox': TFDetector.__convert_coords(b)
                    }
                    detections_cur_image.append(detection_entry)
                    if s > max_detection_conf:
                        max_detection_conf = s

            result['max_detection_conf'] = truncate_float(float(max_detection_conf),
                                                          precision=TFDetector.CONF_DIGITS)
            result['detections'] = detections_cur_image

        except Exception as e:
            result['failure'] = TFDetector.FAILURE_TF_INFER
            print('TFDetector: image {} failed during inference: {}'.format(image_id, str(e)))

        return result


#%% Scoring script

class BatchScorer:
    """
    Coordinates scoring images in this Task.

    1. have a synchronized queue that download tasks enqueue and scoring function dequeues - but need to be able to
    limit the size of the queue. We do not want to write the image to disk and then load it in the scoring func.
    """
    def __init__(self, **kwargs):
        print('score.py BatchScorer, __init__()')

        detector_path = kwargs.get('detector_path')
        self.detector = TFDetector(detector_path)

        self.use_url = kwargs.get('use_url')
        if not self.use_url:
            input_container_sas = kwargs.get('input_container_sas')
            self.input_container_client = ContainerClient.from_container_url(input_container_sas)

        self.detection_threshold = kwargs.get('detection_threshold')

        self.image_ids_to_score = kwargs.get('image_ids_to_score')

        # determine if there is metadata attached to each image_id
        self.metadata_available = True if isinstance(self.image_ids_to_score[0], list) else False

    def _download_image(self, image_file) -> Image:
        """
        Args:
            image_file: Public URL if use_url, else the full path from container root

        Returns:
            PIL image loaded
        """
        if not self.use_url:
            downloader = self.input_container_client.download_blob(image_file)
            image_file = io.BytesIO()
            blob_props = downloader.download_to_stream(image_file)

        image = open_image(image_file)
        return image

    def score_images(self) -> list:
        detections = []

        for i in self.image_ids_to_score:

            if self.metadata_available:
                image_id = i[0]
                image_metadata = i[1]
            else:
                image_id = i

            try:
                image = self._download_image(image_id)
            except Exception as e:
                print(f'score.py BatchScorer, score_images, download_image exception: {e}')
                result = {
                    'file': image_id,
                    'failure': TFDetector.FAILURE_IMAGE_OPEN
                }
            else:
                result = self.detector.generate_detections_one_image(
                    image, image_id, detection_threshold=self.detection_threshold)

            if self.metadata_available:
                result['meta'] = image_metadata

            detections.append(result)
            if len(detections) % PRINT_EVERY == 0:
                print(f'scored {len(detections)} images')

        return detections


def main():
    print('score.py, main()')

    # information to determine input and output locations
    api_instance_name = os.environ['API_INSTANCE_NAME']
    job_id = os.environ['AZ_BATCH_JOB_ID']
    task_id = os.environ['AZ_BATCH_TASK_ID']
    mount_point = os.environ['AZ_BATCH_NODE_MOUNTS_DIR']

    # other parameters for the task
    begin_index = int(os.environ['TASK_BEGIN_INDEX'])
    end_index = int(os.environ['TASK_END_INDEX'])

    input_container_sas = os.environ.get('JOB_CONTAINER_SAS', None)  # could be None if use_url
    use_url = os.environ.get('JOB_USE_URL', None)

    if use_url and use_url.lower() == 'true':  # bool of any non-empty string is True
        use_url = True
    else:
        use_url = False

    detection_threshold = float(os.environ['DETECTION_CONF_THRESHOLD'])

    print(f'score.py, main(), api_instance_name: {api_instance_name}, job_id: {job_id}, task_id: {task_id}, '
          f'mount_point: {mount_point}, begin_index: {begin_index}, end_index: {end_index}, '
          f'input_container_sas: {input_container_sas}, use_url (parsed): {use_url}'
          f'detection_threshold: {detection_threshold}')

    job_folder_mounted = os.path.join(mount_point, 'batch-api', f'api_{api_instance_name}', f'job_{job_id}')
    task_out_dir = os.path.join(job_folder_mounted, 'task_outputs')
    os.makedirs(task_out_dir, exist_ok=True)
    task_output_path = os.path.join(task_out_dir, f'job_{job_id}_task_{task_id}.json')

    # test that we can write to output path; also in case there is no image to process
    with open(task_output_path, 'w') as f:
        json.dump([], f)

    # list images to process
    list_images_path = os.path.join(job_folder_mounted, f'{job_id}_images.json')
    with open(list_images_path) as f:
        list_images = json.load(f)
    print(f'score.py, main(), length of list_images: {len(list_images)}')

    if (not isinstance(list_images, list)) or len(list_images) == 0:
        print('score.py, main(), zero images in specified overall list, exiting...')
        sys.exit(0)

    # items in this list can be strings or [image_id, metadata]
    list_images = list_images[begin_index: end_index]
    if len(list_images) == 0:
        print('score.py, main(), zero images in the shard, exiting')
        sys.exit(0)

    print(f'score.py, main(), processing {len(list_images)} images in this Task')

    # model path
    # Path to .pb TensorFlow detector model file, relative to the
    # models/megadetector_copies folder in mounted container
    detector_model_rel_path = os.environ['DETECTOR_REL_PATH']
    detector_path = os.path.join(mount_point, 'models', 'megadetector_copies', detector_model_rel_path)
    assert os.path.exists(detector_path), f'detector is not found at the specified path: {detector_path}'

    # score the images
    scorer = BatchScorer(
        detector_path=detector_path,
        use_url=use_url,
        input_container_sas=input_container_sas,
        detection_threshold=detection_threshold,
        image_ids_to_score=list_images
    )

    try:
        tick = datetime.now()
        detections = scorer.score_images()
        duration = datetime.now() - tick
        print(f'score.py, main(), score_images() duration: {duration}')
    except Exception as e:
        raise RuntimeError(f'score.py, main(), exception in score_images(): {e}')

    with open(task_output_path, 'w', encoding='utf-8') as f:
        json.dump(detections, f, ensure_ascii=False)

if __name__ == '__main__':
    main()



"""

 ubc_to_json.py

 Convert the .csv file provided for the UBC data set to a 
 COCO-camera-traps .json file

 Images were provided in eight folders, each of which contained a .csv
 file with annotations.  Those annotations came in two slightly different 
 formats, the two formats corresponding to folders starting with "SC_" and 
 otherwise.

"""

#%% Constants and environment

import pandas as pd
import os
import json
import uuid
import numpy as np
import shutil

from tqdm import tqdm
from PIL import Image

from megadetector.visualization import visualize_db
from megadetector.data_management.databases import integrity_check_json_db
from megadetector.utils.path_utils import find_images

input_base = r'e:\ubc'
assert(os.path.isdir(input_base))

output_base = r'f:\data_staging\ubc'
output_json_file = os.path.join(output_base,'ubc.json')
file_list_file = os.path.join(output_base,'all_files.txt')

os.makedirs(output_base,exist_ok=True)

# Map Excel column names - which vary a little across spreadsheets - to a common set of names
mapped_fields = {"Survey.Name" : "survey_name",
                 "project_id": "survey_name",
                 "Camera.Name": "camera_name",
                 "station_id": "camera_name",
                 "Media.Filename": "filename",
                 "orig_file": "filename",
                 "timestamp_pst": "datetime",
                 "Date.Time": "datetime",
                 "Species": "species",
                 "latin_name": "species",
                 "common.name": "common_name",
                 "common_names": "common_name",
                 "Sighting.Quantity": "species_count"
                }

category_mappings = {
    'bird_spp.':'unknown_bird',
    'dog_dog':'dog',
    'hiker_hiker':'hiker',
    'quad_quad':'quad',
    'skier_skier':'skier',
    'snowshoer_snowshoer':'showshoer',
    'quad_quad':'quad'
    }

target_fields = ['species_count','group_count','behaviour']


#%% Enumerate images

# Load from file if we've already enumerated
if os.path.isfile(file_list_file):
    with open(file_list_file,'r') as f:
        files = f.readlines()
    files = [s.strip() for s in files]
    image_full_paths = files
    print('Loaded {} images from {}'.format(len(image_full_paths),file_list_file))
else:
    image_full_paths = find_images(input_base, recursive=True)
    with open(file_list_file,'w') as f:
        for line in image_full_paths:
            f.write(line + '\n')        
    print('Enumerated {} images from {}'.format(len(image_full_paths),input_base))
    
image_full_paths_set = set(image_full_paths)
image_relative_paths = [os.path.relpath(fn,input_base) for fn in image_full_paths]
image_relative_paths_set = set(image_relative_paths)


#%% Create CCT dictionaries

annotations = []
image_ids_to_images = {}
category_name_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
empty_category['common_name'] = 'empty'
category_name_to_category['empty'] = empty_category
next_category_id = 1

latin_to_common = {}

folders = os.listdir(input_base)

# To simplify debugging of the loop below
i_folder = 0; folder = folders[i_folder]; 


##%% Create CCT dictionaries (loop)

invalid_images = []

for i_folder,folder in enumerate(folders):
        
    ##%% 
    
    print('\nProcessing folder {} of {}: {}'.format(i_folder,len(folders),folder))
    
    filenames_to_rows = {}
    filenames_with_multiple_annotations = []
    missing_images = []
    image_directory = os.path.join(input_base, folder)
    files = os.listdir(image_directory)
    files = list(filter(lambda f: f.endswith('.csv'), files))
    input_metadata_file = os.path.join(input_base, folder, files[0])
    assert(os.path.isfile(input_metadata_file))
    
    # Read source data for this folder
    input_metadata = pd.read_csv(input_metadata_file)
    
    # Rename columns
    input_metadata.rename(columns=mapped_fields, inplace=True)
    print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns), len(input_metadata)))
    
    if folder.startswith("SC_"):
        # Folder name is the first two characters of the filename
        #
        # Create relative path names from the filename itself
        input_metadata['image_relative_path'] = input_metadata['filename'].apply(
            lambda x: os.path.join(folder, x[0:2], x.replace(".JPG", ".jpg")))
    else:
        # Folder name is the camera name
        #
        # Create relative path names from camera name and filename    
        input_metadata['image_relative_path'] = input_metadata[['camera_name', 'filename']].apply(
            lambda x: os.path.join(folder, x[0], x[1]), axis = 1)

    # Which of our images are in the spreadsheet?
    # i_row = 0; fn = input_metadata['image_relative_path'][i_row]        
    for i_row, image_relative_path in enumerate(input_metadata['image_relative_path']):
        
        if image_relative_path in filenames_to_rows:
            filenames_with_multiple_annotations.append(image_relative_path)
            filenames_to_rows[image_relative_path].append(i_row)
        else:
            filenames_to_rows[image_relative_path] = [i_row]
            image_full_path = os.path.join(input_base, image_relative_path)
            
            if not image_full_path in image_full_paths_set:
                assert image_relative_path not in image_relative_paths_set
                missing_images.append(image_full_path)
        
    print('Finished verifying image existence for {} files in {} rows\nFound {} filenames with multiple labels, {} missing images'.format(
        len(input_metadata), len(filenames_to_rows), 
        len(filenames_with_multiple_annotations), len(missing_images)))
    
    
    ##%% Check for images that aren't included in the metadata file

    # Find all the images in this folder
    image_relative_paths_this_folder = [s for s in image_relative_paths if s.startswith(folder)]

    # Which of these aren't in the spreadsheet?
    annotated_files_this_folder = list(filenames_to_rows.keys())
    annotated_files_this_folder_set = set(annotated_files_this_folder)
    unannotated_images = [s for s in image_relative_paths_this_folder if s not in annotated_files_this_folder_set]

    print('Found {} unannotated images (of {}) in this folder'.format(
            len(unannotated_images),len(image_relative_paths_this_folder)))
    
    
    ##%% Create entries in CCT dictionaries
    
    image_relative_path = list(filenames_to_rows.keys())[0]
    
    for image_relative_path in list(filenames_to_rows.keys()):
        
        # Only process images we have on disk
        if image_relative_path not in image_relative_paths_set:
            continue
        
        image_full_path = os.path.join(input_base,image_relative_path)
        
        # This is redundant, but doing this for clarity, at basically no performance
        # cost since we need to *read* the images below to check validity.
        assert os.path.isfile(image_full_path)
        
        img_id = image_relative_path.replace('\\','/').replace('/','_').replace(' ','_')
        row_indices = filenames_to_rows[image_relative_path]
        
        # i_row = row_indices[0]
        for i_row in row_indices:
            
            row = input_metadata.iloc[i_row]
            assert(row['image_relative_path'] == image_relative_path)
            timestamp = row['datetime']            
            location = row['survey_name'] + '_' + row['camera_name']
            
            if img_id in image_ids_to_images:
                im = image_ids_to_images[img_id]
                assert im['file_name'] == image_relative_path
                assert im['location'] == location
            else:
                im = {}
                
                try:
                    pil_image = Image.open(image_full_path)        
                    width, height = pil_image.size
                    im['width'] = width
                    im['height'] = height
                except:
                    # These generally represent zero-byte images in this data set, don't try
                    # to find the very small handful that might be other kinds of failures we 
                    # might want to keep around.
                    # print('Error opening image {}'.format(image_relative_path))
                    invalid_images.append(image_relative_path)
                    continue
                
                im['id'] = img_id
                im['file_name'] = image_relative_path
                im['datetime'] = timestamp
                im['location'] = location

                image_ids_to_images[img_id] = im

            species = row['species'].lower().strip().replace(' ','_')
            
            if (isinstance(species,float) or \
                (isinstance(species,str) and (len(species) == 0))):
                category_name = 'empty'
            else:
                category_name = species
            del species
            
            category_name = category_name.strip().lower()
            
            common_name = row['common_name']
            if isinstance(common_name,float) and np.isnan(common_name):
                common_name = ''
            else:
                common_name = str(common_name).lower().strip().replace(', ',',').replace(' ','_')
            
            for k,v in category_mappings.items():
                common_name = common_name.replace(k,v)
                category_name = category_name.replace(k,v)
            common_name = common_name.replace('.','').replace('spp','species')
            category_name = category_name.replace('.','').replace('spp','species')
                        
            if category_name == 'passerine_species' and common_name != '' and common_name != 'passerine_species':
                category_name = common_name
                
            # If we've seen this category before...
            if category_name in category_name_to_category:
                
                category = category_name_to_category[category_name]
                category_id = category['id'] 
                
                # ...make sure it used the same latin --> common mapping
                #
                # If the previous instance had no mapping, use the new one.
                if category['common_name'] == '':
                    category['common_name'] = common_name
                else:
                    # assert common_name == category['common_name']
                    if common_name != category['common_name']:
                        print('Warning: common name {} used for species {}, previously {}'.format(
                            common_name,category_name,category['common_name']))
                    
            else:
                
                category_id = next_category_id
                category = {}
                category['id'] = category_id
                category['name'] = category_name
                category['common_name'] = common_name
                category_name_to_category[category_name] = category
                next_category_id += 1
            
            # Create an annotation
            ann = {}        
            ann['id'] = str(uuid.uuid1())
            ann['image_id'] = im['id']    
            ann['category_id'] = category_id
                        
            for target_field in target_fields:
                if target_field in input_metadata.columns:
                    val = row[target_field]
                    if isinstance(val,float) and np.isnan(val):
                        val = ''
                    else:
                        val = str(val).strip()
                    ann[target_field] = val
                
            annotations.append(ann)

        # ...for each annotation we found for this image
            
    # ...for each image
             
# ...for each dataset
            
images = list(image_ids_to_images.values())
categories = list(category_name_to_category.values())

# Print all of our species mappings
for c in categories:
    print(c['name'].ljust(30) + c['common_name'])

print('Finished creating CCT dictionaries, loaded {} images of {} total on disk ({} invalid)'.format(
    len(images), len(image_relative_paths_set), len(invalid_images)))


#%% Copy images for which we actually have annotations to a new folder, lowercase everything

# im = images[0]
for im in tqdm(images):
    relative_filename = im['file_name']
    input_filename = os.path.join(input_base,relative_filename)
    output_filename = os.path.join(output_base,relative_filename).lower()
    os.makedirs(os.path.dirname(output_filename),exist_ok=True)
    
    shutil.copy(input_filename, output_filename)
    im['file_name'] = im['file_name'].lower()
    im['id'] = im['id'].lower()
    

#%% Create info struct

info = {}
info['year'] = 2020
info['version'] = 1
info['description'] = 'UBC Camera Traps'
info['contributor'] = 'UBC'


#%% Convert image IDs to lowercase in annotations, tag as sequence level

# While there isn't any sequence information, the nature of false positives
# here leads me to believe the images were labeled at the sequence level, so
# we should trust labels more when positives are verified.  Overall false
# positive rate looks to be between 1% and 5%.
    
for ann in annotations:
    ann['image_id'] = ann['image_id'].lower()
    ann['sequence_level_annotation'] = True


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_json_file, 'w'), indent=2)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
    len(images), len(annotations), len(categories)))


#%% Validate output

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = output_base
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = True

sortedCategories, data, errors = integrity_check_json_db.integrity_check_json_db(
    output_json_file, options)


#%% Preview labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 2000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file, image_db = visualize_db.visualize_db(db_path=output_json_file,
                                                         output_dir=os.path.join(
                                                             output_base, 'preview'),
                                                         image_base_dir=output_base,
                                                         options=viz_options)
os.startfile(html_output_file)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import string
import uuid
import threading
from datetime import timedelta

import sas_blob_utils
from flask import Flask, request, jsonify

import server_api_config as api_config
from server_app_config import AppConfig
from server_batch_job_manager import BatchJobManager
from server_orchestration import create_batch_job, monitor_batch_job
from server_job_status_table import JobStatusTable
from server_utils import *

# %% Flask app
app = Flask(__name__)

# reference: https://trstringer.com/logging-flask-gunicorn-the-manageable-way/
if __name__ != '__main__':
    gunicorn_logger = logging.getLogger('gunicorn.error')
    app.logger.handlers = gunicorn_logger.handlers
    app.logger.setLevel(gunicorn_logger.level)


API_PREFIX = api_config.API_PREFIX
app.logger.info('server, created Flask application...')

# %% Helper classes

app_config = AppConfig()
job_status_table = JobStatusTable()
batch_job_manager = BatchJobManager()
app.logger.info('server, finished instantiating helper classes')


# %% Flask endpoints

@app.route(f'{API_PREFIX}/')
def hello():
    return f'Camera traps batch processing API. Instance: {api_config.API_INSTANCE_NAME}'


@app.route(f'{API_PREFIX}/request_detections', methods=['POST'])
def request_detections():
    """
    Checks that the input parameters to this endpoint are valid, starts a thread
    to launch the batch processing job, and return the job_id/request_id to the user.
    """
    if not request.is_json:
        msg = 'Body needs to have a JSON mimetype (e.g., application/json).'
        return make_error(415, msg)

    try:
        post_body = request.get_json()
    except Exception as e:
        return make_error(415, f'Error occurred reading POST request body: {e}.')

    app.logger.info(f'server, request_detections, post_body: {post_body}')

    # required params

    caller_id = post_body.get('caller', None)
    if caller_id is None or caller_id not in app_config.get_allowlist():
        msg = ('Parameter caller is not supplied or is not on our allowlist. '
               'Please email cameratraps@lila.science to request access.')
        return make_error(401, msg)

    use_url = post_body.get('use_url', False)
    if use_url and isinstance(use_url, str):  # in case it is included but is intended to be False
        if use_url.lower() in ['false', 'f', 'no', 'n']:
            use_url = False

    input_container_sas = post_body.get('input_container_sas', None)
    if not input_container_sas and not use_url:
        msg = ('input_container_sas with read and list access is a required '
               'field when not using image URLs.')
        return make_error(400, msg)

    if input_container_sas is not None:
        if not sas_blob_utils.is_container_uri(input_container_sas):
            return make_error(400, 'input_container_sas provided is not for a container.')

        result = check_data_container_sas(input_container_sas)
        if result is not None:
            return make_error(result[0], result[1])

    # can be an URL to a file not hosted in an Azure blob storage container
    images_requested_json_sas = post_body.get('images_requested_json_sas', None)

    if images_requested_json_sas is not None:
        if not images_requested_json_sas.startswith(('http://', 'https://')):
            return make_error(400, 'images_requested_json_sas needs to be an URL.')

    # if use_url, then images_requested_json_sas is required
    if use_url and images_requested_json_sas is None:
            return make_error(400, 'images_requested_json_sas is required since use_url is true.')

    # optional params

    # check model_version is among the available model versions
    model_version = post_body.get('model_version', '')
    if model_version != '':
        model_version = str(model_version)  # in case user used an int
        if model_version not in api_config.MD_VERSIONS_TO_REL_PATH:
            return make_error(400, f'model_version {model_version} is not supported.')

    # check request_name has only allowed characters
    request_name = post_body.get('request_name', '')
    if request_name != '':
        if len(request_name) > 92:
            return make_error(400, 'request_name is longer than 92 characters.')
        allowed = set(string.ascii_letters + string.digits + '_' + '-')
        if not set(request_name) <= allowed:
            msg = ('request_name contains invalid characters (only letters, '
                   'digits, - and _ are allowed).')
            return make_error(400, msg)

    # optional params for telemetry collection - logged to status table for now as part of call_params
    country = post_body.get('country', None)
    organization_name = post_body.get('organization_name', None)

    # All API instances / node pools share a quota on total number of active Jobs;
    # we cannot accept new Job submissions if we are at the quota
    try:
        num_active_jobs = batch_job_manager.get_num_active_jobs()
        if num_active_jobs >= api_config.MAX_BATCH_ACCOUNT_ACTIVE_JOBS:
            return make_error(503, f'Too many active jobs, please try again later')
    except Exception as e:
        return make_error(500, f'Error checking number of active jobs: {e}')

    try:
        job_id = uuid.uuid4().hex
        job_status_table.create_job_status(
            job_id=job_id,
            status= get_job_status('created', 'Request received. Listing images next...'),
            call_params=post_body
        )
    except Exception as e:
        return make_error(500, f'Error creating a job status entry: {e}')

    try:
        thread = threading.Thread(
            target=create_batch_job,
            name=f'job_{job_id}',
            kwargs={'job_id': job_id, 'body': post_body}
        )
        thread.start()
    except Exception as e:
        return make_error(500, f'Error creating or starting the batch processing thread: {e}')

    return {'request_id': job_id}


@app.route(f'{API_PREFIX}/cancel_request', methods=['POST'])
def cancel_request():
    """
    Cancels a request/job given the job_id and caller_id
    """
    if not request.is_json:
        msg = 'Body needs to have a JSON mimetype (e.g., application/json).'
        return make_error(415, msg)
    try:
        post_body = request.get_json()
    except Exception as e:
        return make_error(415, f'Error occurred reading POST request body: {e}.')

    app.logger.info(f'server, cancel_request received, body: {post_body}')

    # required fields
    job_id = post_body.get('request_id', None)
    if job_id is None:
        return make_error(400, 'request_id is a required field.')

    caller_id = post_body.get('caller', None)
    if caller_id is None or caller_id not in app_config.get_allowlist():
        return make_error(401, 'Parameter caller is not supplied or is not on our allowlist.')

    item_read = job_status_table.read_job_status(job_id)
    if item_read is None:
        return make_error(404, 'Task is not found.')
    if 'status' not in item_read:
        return make_error(404, 'Something went wrong. This task does not have a status field.')

    request_status = item_read['status']['request_status']
    if request_status not in ['running', 'problem']:
        # request_status is either completed or failed
        return make_error(400, f'Task has {request_status} and cannot be canceled')

    try:
        batch_job_manager.cancel_batch_job(job_id)
        # the create_batch_job thread will stop when it wakes up the next time
    except Exception as e:
        return make_error(500, f'Error when canceling the request: {e}')
    else:
        job_status_table.update_job_status(job_id, {
            'request_status': 'canceled',
            'message': 'Request has been canceled by the user.'
        })
    return 'Canceling signal has been sent. You can verify the status at the /task endpoint'


@app.route(f'{API_PREFIX}/task/<job_id>')
def retrieve_job_status(job_id: str):
    """
    Does not require the "caller" field to avoid checking the allowlist in App Configurations.
    Retains the /task endpoint name to be compatible with previous versions.
    """
    # Fix for Zooniverse - deleting any "-" characters in the job_id
    job_id = job_id.replace('-', '')

    item_read = job_status_table.read_job_status(job_id)  # just what the monitoring thread wrote to the DB
    if item_read is None:
        return make_error(404, 'Task is not found.')
    if 'status' not in item_read or 'last_updated' not in item_read or 'call_params' not in item_read:
        return make_error(404, 'Something went wrong. This task does not have a valid status.')

    # If the status is running, it could be a Job submitted before the last restart of this
    # API instance. If that is the case, we should start to monitor its progress again.
    status = item_read['status']

    last_updated = datetime.fromisoformat(item_read['last_updated'][:-1])  # get rid of "Z" (required by Cosmos DB)
    time_passed = datetime.utcnow() - last_updated
    job_is_unmonitored = True if time_passed > timedelta(minutes=(api_config.MONITOR_PERIOD_MINUTES + 1)) else False

    if isinstance(status, dict) and \
            'request_status' in status and \
            status['request_status'] in ['running', 'problem'] and \
            'num_tasks' in status and \
            job_id not in get_thread_names() and \
            job_is_unmonitored:
        # WARNING model_version could be wrong (a newer version number gets written to the output file) around
        # the time that  the model is updated, if this request was submitted before the model update
        # and the API restart; this should be quite rare
        model_version = item_read['call_params'].get('model_version', api_config.DEFAULT_MD_VERSION)

        num_tasks = status['num_tasks']
        job_name = item_read['call_params'].get('request_name', '')
        job_submission_timestamp = item_read.get('job_submission_time', '')

        thread = threading.Thread(
            target=monitor_batch_job,
            name=f'job_{job_id}',
            kwargs={
                'job_id': job_id,
                'num_tasks': num_tasks,
                'model_version': model_version,
                'job_name': job_name,
                'job_submission_timestamp': job_submission_timestamp
            }
        )
        thread.start()
        app.logger.info(f'server, started a new thread to monitor job {job_id}')

    # conform to previous schemes
    if 'num_tasks' in status:
        del status['num_tasks']
    item_to_return = {
        'Status': status,
        'Endpoint': f'{API_PREFIX}/request_detections',
        'TaskId': job_id,
        'Timestamp': item_read['last_updated']
    }
    return item_to_return


@app.route(f'{API_PREFIX}/default_model_version')
def get_default_model_version() -> str:
    return api_config.DEFAULT_MD_VERSION


@app.route(f'{API_PREFIX}/supported_model_versions')
def get_supported_model_versions() -> str:
    return jsonify(sorted(list(api_config.MD_VERSIONS_TO_REL_PATH.keys())))


# %% undocumented endpoints

def get_thread_names() -> list:
    thread_names = []
    for thread in threading.enumerate():
        if thread.name.startswith('job_'):
            thread_names.append(thread.name.split('_')[1])
    return sorted(thread_names)


@app.route(f'{API_PREFIX}/all_jobs')
def get_all_jobs():
    """List all Jobs being monitored since this API instance started"""
    thread_names = get_thread_names()
    return jsonify(thread_names)



"""

test_lila_metadata_urls.py

Test that all the metadata URLs for LILA camera trap datasets are valid, including MegaDetector 
results files.

Also pick an arbitrary image from each dataset and make sure that URL is valid.

Also picks an arbitrary image from each dataset's MD results and make sure the corresponding URL is valid.

"""

#%% Constants and imports

import json
import os

from megadetector.data_management.lila.lila_common import read_lila_metadata,\
    read_metadata_file_for_dataset, read_lila_taxonomy_mapping

# We'll write images, metadata downloads, and temporary files here
lila_local_base = os.path.expanduser('~/lila')

output_dir = os.path.join(lila_local_base,'lila_metadata_tests')
os.makedirs(output_dir,exist_ok=True)

metadata_dir = os.path.join(lila_local_base,'metadata')
os.makedirs(metadata_dir,exist_ok=True)

md_results_dir = os.path.join(lila_local_base,'md_results')
os.makedirs(md_results_dir,exist_ok=True)

md_results_keys = ['mdv4_results_raw','mdv5a_results_raw','mdv5b_results_raw','md_results_with_rde']

preferred_cloud = 'gcp' # 'azure', 'aws'


#%% Load category and taxonomy files

taxonomy_df = read_lila_taxonomy_mapping(metadata_dir)


#%% Download and parse the metadata file

metadata_table = read_lila_metadata(metadata_dir)

print('Loaded metadata URLs for {} datasets'.format(len(metadata_table)))


#%% Download and extract metadata and MD results for each dataset

for ds_name in metadata_table.keys():    

    metadata_table[ds_name]['json_filename'] = read_metadata_file_for_dataset(ds_name=ds_name,
                                                                         metadata_dir=metadata_dir,
                                                                         metadata_table=metadata_table)
    for k in md_results_keys:
        md_results_url = metadata_table[ds_name][k]
        if md_results_url is None:
            metadata_table[ds_name][k + '_filename'] = None
        else:
            metadata_table[ds_name][k + '_filename'] = read_metadata_file_for_dataset(ds_name=ds_name,
                                                                        metadata_dir=md_results_dir,
                                                                        json_url=md_results_url)


#%% Build up a list of URLs to test

# Takes ~15 mins, since it has to open all the giant .json files

url_to_source = {}

# The first image in a dataset is disproportionately likely to be human (and thus 404),
# so we pick a semi-arbitrary image that isn't the first.  How about the 1000th?
image_index = 1000

# ds_name = list(metadata_table.keys())[0]
for ds_name in metadata_table.keys():
    
    if 'bbox' in ds_name:
        print('Skipping bbox dataset {}'.format(ds_name))
        continue

    print('Processing dataset {}'.format(ds_name))
    
    json_filename = metadata_table[ds_name]['json_filename']
    with open(json_filename, 'r') as f:
        data = json.load(f)

    image_base_url = metadata_table[ds_name]['image_base_url_' + preferred_cloud]
    assert not image_base_url.endswith('/')
    # Download a test image
    test_image_relative_path = data['images'][image_index]['file_name']
    test_image_url = image_base_url + '/' + test_image_relative_path
    
    url_to_source[test_image_url] = ds_name + ' metadata'
    
    # Grab an image from the MegaDetector results
    
    # k = md_results_keys[2]
    for k in md_results_keys:
        k_fn = k + '_filename'
        if metadata_table[ds_name][k_fn] is not None:
            with open(metadata_table[ds_name][k_fn],'r') as f:
                md_results = json.load(f)
                im = md_results['images'][image_index]
                md_image_url = image_base_url + '/' + im['file']
                url_to_source[md_image_url] = ds_name + ' ' + k
            del md_results
    del data
        
# ...for each dataset


#%% Test URLs

from megadetector.utils.url_utils import test_urls

urls_to_test = sorted(url_to_source.keys())
urls_to_test = [fn.replace('\\','/') for fn in urls_to_test]

status_codes = test_urls(urls_to_test,
                         error_on_failure=False,
                         pool_type='thread',
                         n_workers=10,
                         timeout=2.0)

for i_url,url in enumerate(urls_to_test):
    if status_codes[i_url] != 200:
        print('Status {} for {} ({})'.format(
            status_codes[i_url],url,url_to_source[url]))



"""

remove_repeat_detections.py

Used after running find_repeat_detections, then manually filtering the results,
to create a final filtered output file.

If you want to use this script, we recommend that you read the RDE user's guide:

https://github.com/agentmorris/MegaDetector/tree/main/megadetector/postprocessing/repeat_detection_elimination

"""

#%% Constants and imports

import argparse
import os

from megadetector.postprocessing.repeat_detection_elimination import repeat_detections_core


#%% Main function

def remove_repeat_detections(inputFile,outputFile,filteringDir):
    """
    Given an index file that was produced in a first pass through find_repeat_detections,
    and a folder of images (from which the user has deleted images they don't want removed),
    remove the identified repeat detections from a set of MD results and write to a new file.
    
    Args:
        inputFile (str): .json file of MD results, from which we should remove repeat detections
        outputFile (str): output .json file to which we should write MD results (with repeat 
            detections removed)
        filteringDir (str): the folder produced by find_repeat_detections, containing a 
            detectionIndex.json file        
    """
    
    assert os.path.isfile(inputFile), "Can't find file {}".format(inputFile)
    assert os.path.isdir(filteringDir), "Can't find folder {}".format(filteringDir)
    options = repeat_detections_core.RepeatDetectionOptions()
    if os.path.isfile(filteringDir):
        options.filterFileToLoad = filteringDir
    else:
        assert os.path.isdir(filteringDir), '{} is not a valid folder'.format(filteringDir)
        options.filterFileToLoad = \
            os.path.join(filteringDir,repeat_detections_core.detection_index_file_name_base)
    repeat_detections_core.find_repeat_detections(inputFile, outputFile, options)


#%% Interactive driver

if False:
    
    #%%
    
    inputFile = r''
    outputFile = r''
    filteringDir = r''
    remove_repeat_detections(inputFile,outputFile,filteringDir)


#%% Command-line driver

import sys

def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('inputFile', help='.json file containing the original, unfiltered API results')
    parser.add_argument('outputFile', help='.json file to which you want to write the final, ' + \
                        'filtered API results')
    parser.add_argument('filteringDir', help='directory where you looked at lots of images and ' + \
                        'decided which ones were really false positives')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    remove_repeat_detections(args.inputFile, args.outputFile, args.filteringDir)

if __name__ == '__main__':
    main()



"""

 add_timestamps_to_icct.py

 The Island Conservation Camera Traps dataset was originally posted without timestamps
 in either .json metadata or EXIF metadata.  We pulled timestamps out using ocr_tools.py,
 this script adds those timestamps into the .json metadata.

"""

#%% Imports and constants

import json

ocr_results_file = r'g:\temp\ocr_results.2023.10.31.07.37.54.json'
input_metadata_file = r'd:\lila\islandconservationcameratraps\island_conservation.json'
output_metadata_file = r'g:\temp\island_conservation_camera_traps_1.02.json'
ocr_results_file_base = 'g:/temp/island_conservation_camera_traps/'
assert ocr_results_file_base.endswith('/')


#%% Read input metadata

with open(input_metadata_file,'r') as f:
    input_metadata = json.load(f)
    
assert input_metadata['info']['version'] == '1.01'

# im = input_metadata['images'][0]
for im in input_metadata['images']:
    assert 'datetime' not in im
    

#%% Read OCR results

with open(ocr_results_file,'r') as f:
    abs_filename_to_ocr_results = json.load(f)
    
relative_filename_to_ocr_results = {}

for fn_abs in abs_filename_to_ocr_results:
    assert ocr_results_file_base in fn_abs
    fn_relative = fn_abs.replace(ocr_results_file_base,'')
    relative_filename_to_ocr_results[fn_relative] = abs_filename_to_ocr_results[fn_abs]
    

#%% Add datetimes to metadata

images_not_in_datetime_results = []
images_with_failed_datetimes = []

for i_image,im in enumerate(input_metadata['images']):
    if im['file_name'] not in relative_filename_to_ocr_results:
        images_not_in_datetime_results.append(im)
        im['datetime'] = None
        continue
    ocr_results = relative_filename_to_ocr_results[im['file_name']]
    if ocr_results['datetime'] is None:
        images_with_failed_datetimes.append(im)
        im['datetime'] = None
        continue
    im['datetime'] = ocr_results['datetime']

print('{} of {} images were not in datetime results'.format(
    len(images_not_in_datetime_results),len(input_metadata['images'])))

print('{} of {} images were had failed datetime results'.format(
    len(images_with_failed_datetimes),len(input_metadata['images'])))

for im in input_metadata['images']:
    assert 'datetime' in im


#%% Write output

input_metadata['info']['version'] = '1.02'

with open(output_metadata_file,'w') as f:
    json.dump(input_metadata,f,indent=1)


"""

add_locations_to_island_camera_traps.py

The Island Conservation Camera Traps dataset had unique camera identifiers embedded
in filenames, but not in the proper metadata fields.  This script copies that information
to metadata.

"""

#%% Imports and constants

import os
import json
from tqdm import tqdm

input_fn = os.path.expanduser('~/lila/metadata/island_conservation.json')
output_fn = os.path.expanduser('~/tmp/island_conservation.json')
preview_folder = os.path.expanduser('~/tmp/island_conservation_preview')
image_directory = os.path.expanduser('~/data/icct/public/')


#%% Read input file

with open(input_fn,'r') as f:
    d = json.load(f)
    
d['info']
d['info']['version'] = '1.01'


#%% Find locations

images = d['images']

locations = set()

for i_image,im in tqdm(enumerate(images),total=len(images)):
    tokens_fn = im['file_name'].split('/')
    tokens_id = im['id'].split('_')
    assert tokens_fn[0] == tokens_id[0]
    assert tokens_fn[1] == tokens_id[1]
    location = tokens_fn[0] + '_' + tokens_fn[1]
    im['location'] = location
    locations.add(location)

locations = sorted(list(locations))
    
for s in locations:
    print(s)
    
    
#%% Write output file

with open(output_fn,'w') as f:
    json.dump(d,f,indent=1)
    

#%% Validate .json files

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_directory
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = True

sorted_categories, data, error_info = integrity_check_json_db.integrity_check_json_db(output_fn, options)


#%% Preview labels

from megadetector.visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 2000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['test']
html_output_file, image_db = visualize_db.visualize_db(db_path=output_fn,
                                                         output_dir=preview_folder,
                                                         image_base_dir=image_directory,
                                                         options=viz_options)

from megadetector.utils import path_utils
path_utils.open_file(html_output_file)


#%% Zip output file

from megadetector.utils.path_utils import zip_file

zip_file(output_fn, verbose=True)
assert os.path.isfile(output_fn + '.zip')



"""

convert_output_format.py

Converts between file formats output by our batch processing API.  Currently
supports json <--> csv conversion, but this should be the landing place for any
conversion - including between hypothetical alternative .json versions - that we support 
in the future.

The .csv format is largely obsolete, don't use it unless you're super-duper sure you need it.

"""

#%% Constants and imports

import argparse
import json
import csv
import sys
import os

from tqdm import tqdm

from megadetector.postprocessing.load_api_results import load_api_results_csv
from megadetector.data_management.annotations import annotation_constants
from megadetector.utils import ct_utils

CONF_DIGITS = 3


#%% Conversion functions

def convert_json_to_csv(input_path,
                        output_path=None,
                        min_confidence=None,
                        omit_bounding_boxes=False,
                        output_encoding=None,
                        overwrite=True):
    """
    Converts a MD results .json file to a totally non-standard .csv format.
    
    If [output_path] is None, will convert x.json to x.csv.
    
    TODO: this function should obviously be using Pandas or some other sensible structured
    representation of tabular data.  Even a list of dicts.  This implementation is quite
    brittle and depends on adding fields to every row in exactly the right order.
    
    Args:
        input_path (str): the input .json file to convert
        output_path (str, optional): the output .csv file to generate; if this is None, uses 
            [input_path].csv
        min_confidence (float, optional): the minimum-confidence detection we should include 
            in the "detections" column; has no impact on the other columns
        omit_bounding_boxes (bool): whether to leave out the json-formatted bounding boxes
            that make up the "detections" column, which are not generally useful for someone who
            wants to consume this data as a .csv file
        output_encoding (str, optional): encoding to use for the .csv file
        overwrite (bool): whether to overwrite an existing .csv file; if this is False and the
            output file exists, no-ops and returns
           
    """
    
    if output_path is None:
        output_path = os.path.splitext(input_path)[0]+'.csv'
        
    if os.path.isfile(output_path) and (not overwrite):
        print('File {} exists, skipping json --> csv conversion'.format(output_path))
        return
    
    print('Loading json results from {}...'.format(input_path))
    json_output = json.load(open(input_path))

    rows = []
    
    fixed_columns = ['image_path', 'max_confidence', 'detections']
    
    # We add an output column for each class other than 'empty', 
    # containing the maximum probability of  that class for each image
    # n_non_empty_detection_categories = len(annotation_constants.annotation_bbox_categories) - 1
    n_non_empty_detection_categories = annotation_constants.NUM_DETECTOR_CATEGORIES
    detection_category_column_names = []
    assert annotation_constants.detector_bbox_category_id_to_name[0] == 'empty'
    for cat_id in range(1,n_non_empty_detection_categories+1):
        cat_name = annotation_constants.detector_bbox_category_id_to_name[cat_id]
        detection_category_column_names.append('max_conf_' + cat_name)
    
    n_classification_categories = 0
    
    if 'classification_categories' in json_output.keys():
        classification_category_id_to_name = json_output['classification_categories']
        classification_category_ids = list(classification_category_id_to_name.keys())
        classification_category_id_to_column_number = {}
        classification_category_column_names = []
        for i_category,category_id in enumerate(classification_category_ids):
            category_name = classification_category_id_to_name[category_id].\
                replace(' ','_').replace(',','')
            classification_category_column_names.append('max_classification_conf_' + category_name)
            classification_category_id_to_column_number[category_id] = i_category

        n_classification_categories = len(classification_category_ids)
    
    # There are several .json fields for which we add .csv columns; other random bespoke fields
    # will be ignored.
    optional_fields = ['width','height','datetime','exif_metadata']
    optional_fields_present = set()
    
    # Iterate once over the data to check for optional fields
    print('Looking for optional fields...')
    
    for im in tqdm(json_output['images']):
        # Which optional fields are present for this image?
        for k in im.keys():
            if k in optional_fields:
                optional_fields_present.add(k)
                
    optional_fields_present = sorted(list(optional_fields_present))
    if len(optional_fields_present) > 0:
        print('Found {} optional fields'.format(len(optional_fields_present)))
            
    expected_row_length = len(fixed_columns) + len(detection_category_column_names) + \
        n_classification_categories + len(optional_fields_present)
        
    print('Formatting results...')
    
    # i_image = 0; im = json_output['images'][i_image]
    for im in tqdm(json_output['images']):
        
        image_id = im['file']
        
        if 'failure' in im and im['failure'] is not None:
            row = [image_id, 'failure', im['failure']]
            rows.append(row)
            # print('Skipping failed image {} ({})'.format(im['file'],im['failure']))
            continue

        max_conf = ct_utils.get_max_conf(im)
        detections = []
        max_detection_category_probabilities = [None] * n_non_empty_detection_categories
        max_classification_category_probabilities = [0] * n_classification_categories
              
        # d = im['detections'][0]
        for d in im['detections']:
            
            # Skip sub-threshold detections
            if (min_confidence is not None) and (d['conf'] < min_confidence):
                continue
            
            input_bbox = d['bbox']
            
            # Our .json format is xmin/ymin/w/h
            #
            # Our .csv format was ymin/xmin/ymax/xmax
            xmin = input_bbox[0]
            ymin = input_bbox[1]
            xmax = input_bbox[0] + input_bbox[2]
            ymax = input_bbox[1] + input_bbox[3]
            output_detection = [ymin, xmin, ymax, xmax]
                
            output_detection.append(d['conf'])
            
            # Category 0 is empty, for which we don't have a column, so the max
            # confidence for category N goes in column N-1
            detection_category_id = int(d['category'])
            assert detection_category_id > 0 and detection_category_id <= \
                n_non_empty_detection_categories
            detection_category_column = detection_category_id - 1
            detection_category_max = max_detection_category_probabilities[detection_category_column]
            if detection_category_max is None or d['conf'] > detection_category_max:
                max_detection_category_probabilities[detection_category_column] = d['conf']
            
            output_detection.append(detection_category_id)
            detections.append(output_detection)
    
            if 'classifications' in d:
                assert n_classification_categories > 0,\
                    'Oops, I have classification results, but no classification metadata'
                for c in d['classifications']:
                    category_id = c[0]
                    p = c[1]
                    category_index = classification_category_id_to_column_number[category_id]
                    if (max_classification_category_probabilities[category_index] < p):
                        max_classification_category_probabilities[category_index] = p
                
                # ...for each classification
                
            # ...if we have classification results for this detection
            
        # ...for each detection
        
        detection_string = ''
        if not omit_bounding_boxes:
            detection_string = json.dumps(detections)
            
        row = [image_id, max_conf, detection_string]
        row.extend(max_detection_category_probabilities)
        row.extend(max_classification_category_probabilities)
        
        for field_name in optional_fields_present:
            if field_name not in im:
                row.append('')
            else:
                row.append(str(im[field_name]))
                
        assert len(row) == expected_row_length
        rows.append(row)
        
    # ...for each image

    print('Writing to csv...')
    
    with open(output_path, 'w', newline='', encoding=output_encoding) as f:
        writer = csv.writer(f, delimiter=',')
        header = fixed_columns
        header.extend(detection_category_column_names)
        if n_classification_categories > 0:
            header.extend(classification_category_column_names)
        for field_name in optional_fields_present:
            header.append(field_name)
        writer.writerow(header)
        writer.writerows(rows)

# ...def convert_json_to_csv(...)

    
def convert_csv_to_json(input_path,output_path=None,overwrite=True):
    """
    Convert .csv to .json.  If output_path is None, will convert x.csv to x.json.
    
    Args:
        input_path (str): .csv filename to convert to .json
        output_path (str, optional): the output .json file to generate; if this is None, uses 
            [input_path].json
        overwrite (bool): whether to overwrite an existing .json file; if this is False and the
            output file exists, no-ops and returns    
        
    """
    
    if output_path is None:
        output_path = os.path.splitext(input_path)[0]+'.json'
        
    if os.path.isfile(output_path) and (not overwrite):
        print('File {} exists, skipping csv --> json conversion'.format(output_path))
        return
            
    # Format spec:
    #
    # https://github.com/agentmorris/MegaDetector/tree/main/megadetector/api/batch_processing
    
    print('Loading csv results...')
    df = load_api_results_csv(input_path)

    info = {
        "format_version":"1.2",
        "detector": "unknown",
        "detection_completion_time" : "unknown",
        "classifier": "unknown",
        "classification_completion_time": "unknown"
    }
    
    classification_categories = {}
    detection_categories = annotation_constants.detector_bbox_categories

    images = []
    
    # iFile = 0; row = df.iloc[iFile]
    for iFile,row in df.iterrows():
        
        image = {}
        image['file'] = row['image_path']
        image['max_detection_conf'] = round(row['max_confidence'], CONF_DIGITS)
        src_detections = row['detections']        
        out_detections = []
        
        for iDetection,detection in enumerate(src_detections):
            
            # Our .csv format was ymin/xmin/ymax/xmax
            #
            # Our .json format is xmin/ymin/w/h
            ymin = detection[0]
            xmin = detection[1]
            ymax = detection[2]
            xmax = detection[3]
            bbox = [xmin, ymin, xmax-xmin, ymax-ymin]
            conf = detection[4]
            iClass = detection[5]
            out_detection = {}
            out_detection['category'] = str(iClass)
            out_detection['conf'] = conf
            out_detection['bbox'] = bbox            
            out_detections.append(out_detection)
            
        # ...for each detection
        
        image['detections'] = out_detections
        images.append(image)
        
    # ...for each image        
    json_out = {}
    json_out['info'] = info
    json_out['detection_categories'] = detection_categories
    json_out['classification_categories'] = classification_categories
    json_out['images'] = images
    
    json.dump(json_out,open(output_path,'w'),indent=1)
    
# ...def convert_csv_to_json(...)


#%% Interactive driver

if False:    

    #%%
    
    input_path = r'c:\temp\test.json'
    min_confidence = None
    output_path = input_path + '.csv'
    convert_json_to_csv(input_path,output_path,min_confidence=min_confidence,
                        omit_bounding_boxes=False)
            
    #%%
    
    base_path = r'c:\temp\json'
    input_paths = os.listdir(base_path)
    input_paths = [os.path.join(base_path,s) for s in input_paths]
    
    min_confidence = None    
    for input_path in input_paths:
        output_path = input_path + '.csv'
        convert_json_to_csv(input_path,output_path,min_confidence=min_confidence,
                            omit_bounding_boxes=True)    
    
    #%% Concatenate .csv files from a folder

    import glob
    csv_files = glob.glob(os.path.join(base_path,'*.json.csv' ))
    master_csv = os.path.join(base_path,'all.csv')
    
    print('Concatenating {} files to {}'.format(len(csv_files),master_csv))
    
    header = None
    with open(master_csv, 'w') as fout:
        
        for filename in tqdm(csv_files):
            
            with open(filename) as fin:
                
                lines = fin.readlines()
                
                if header is not None:
                    assert lines[0] == header
                else:
                    header = lines[0]
                    fout.write(header)
                    
                for line in lines[1:]:
                    if len(line.strip()) == 0:
                        continue                    
                    fout.write(line)
                    
        # ...for each .csv file
        
    # with open(master_csv)
    
    
#%% Command-line driver
        
def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('input_path',type=str,
                        help='Input filename ending in .json or .csv')
    parser.add_argument('--output_path',type=str,default=None,
                        help='Output filename ending in .json or .csv (defaults to ' + \
                             'input file, with .json/.csv replaced by .csv/.json)')
    parser.add_argument('--omit_bounding_boxes',action='store_true',
                        help='Output bounding box text from .csv output (large and usually not useful)')        
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    
    if args.output_path is None:
        if args.input_path.endswith('.csv'):
            args.output_path = args.input_path[:-4] + '.json'
        elif args.input_path.endswith('.json'):
            args.output_path = args.input_path[:-5] + '.csv'
        else:
            raise ValueError('Illegal input file extension')    
    
    if args.input_path.endswith('.csv') and args.output_path.endswith('.json'):
        assert not args.omit_bounding_boxes, \
            '--omit_bounding_boxes does not apply to csv --> json conversion'
        convert_csv_to_json(args.input_path,args.output_path)
    elif args.input_path.endswith('.json') and args.output_path.endswith('.csv'):
        convert_json_to_csv(args.input_path,args.output_path,omit_bounding_boxes=args.omit_bounding_boxes)
    else:
        raise ValueError('Illegal format combination')            

if __name__ == '__main__':
    main()



"""

map_lila_taxonomy_to_wi_taxonomy.py

Loads the LILA category mapping (in which taxonomy information comes from an 
iNat taxonomy snapshot) and tries to map each class to the Wildlife Insights taxonomy.

"""

#%% Constants and imports

import numpy as np
import json
import os

from tqdm import tqdm

from megadetector.data_management.lila.lila_common import \
    read_lila_taxonomy_mapping, read_wildlife_insights_taxonomy_mapping


#%% Prevent execution during infrastructural imports

if False:
    
    #%%
    
    lila_local_base = os.path.expanduser('~/lila')

    metadata_dir = os.path.join(lila_local_base, 'metadata')
    os.makedirs(metadata_dir, exist_ok=True)

    # Created by get_lila_category_list.py... contains counts for each category
    category_list_dir = os.path.join(lila_local_base, 'lila_categories_list')
    lila_dataset_to_categories_file = os.path.join(
        category_list_dir, 'lila_dataset_to_categories.json')

    # This is a manually-curated file used to store mappings that had to be made manually
    lila_to_wi_supplementary_mapping_file = os.path.expanduser(
        '~/git/MegaDetector/taxonomy_mapping/lila_to_wi_supplementary_mapping_file.csv')

    assert os.path.isfile(lila_dataset_to_categories_file)

    # This is the main output file from this whole process
    wi_mapping_table_file = os.path.join(lila_local_base,'lila_wi_mapping_table.csv')

    id_column = 'uniqueIdentifier' # 'id'


    #%% Load category and taxonomy files

    with open(lila_dataset_to_categories_file, 'r') as f:
        lila_dataset_to_categories = json.load(f)

    lila_taxonomy_df = read_lila_taxonomy_mapping(metadata_dir)

    wi_taxonomy_df = read_wildlife_insights_taxonomy_mapping(metadata_dir)


    #%% Pull everything out of pandas

    lila_taxonomy = lila_taxonomy_df.to_dict('records')
    wi_taxonomy = wi_taxonomy_df.to_dict('records')


    #%% Cache WI taxonomy lookups

    def is_empty_wi_item(v):
        if isinstance(v, str):
            return len(v) == 0        
        elif v is None:
            return True
        else:
            assert isinstance(v, float) and np.isnan(v), 'Invalid item: {}'.format(str(v))
            return True


    def taxonomy_items_equal(a, b):
        if isinstance(a, str) and (not isinstance(b, str)):
            return False
        if isinstance(b, str) and (not isinstance(a, str)):
            return False
        if (not isinstance(a, str)) or (not isinstance(b, str)):
            assert (a is None and b is None) or (isinstance(a, float) and isinstance(b, float))
            return True
        return a == b


    for taxon in wi_taxonomy:
        taxon['taxon_name'] = None

    from collections import defaultdict
    wi_taxon_name_to_taxa = defaultdict(list)

    # This is just a handy lookup table that we'll use to debug mismatches
    wi_common_name_to_taxon = {}

    blank_taxon_name = 'blank'
    blank_taxon = None

    animal_taxon_name = 'animal'
    animal_taxon = None

    unknown_taxon_name = 'unknown'
    unknown_taxon = None

    ignore_taxa = set(['No CV Result', 'CV Needed', 'CV Failed'])

    known_problematic_taxon_ids = ['f94e6d97-59cf-4d38-a05a-a75efdd2863b']

    human_taxa = []

    # taxon = wi_taxonomy[21653]; print(taxon)
    for taxon in tqdm(wi_taxonomy):

        taxon_name = None

        assert taxon['taxonomyType'] == 'object' or taxon['taxonomyType'] == 'biological'

        for k in taxon.keys():
            v = taxon[k]
            if isinstance(v,str):
                taxon[k] = v.strip()
                
        if taxon['commonNameEnglish'] in ignore_taxa:
            continue

        if isinstance(taxon['commonNameEnglish'], str):

            wi_common_name_to_taxon[taxon['commonNameEnglish'].strip(
            ).lower()] = taxon

            special_taxon = False

            # Look for keywords that don't refer to specific taxa: blank/animal/unknown
            if taxon['commonNameEnglish'].strip().lower() == blank_taxon_name:
                blank_taxon = taxon
                special_taxon = True

            elif taxon['commonNameEnglish'].strip().lower() == animal_taxon_name:
                animal_taxon = taxon
                special_taxon = True

            elif taxon['commonNameEnglish'].strip().lower() == unknown_taxon_name:
                unknown_taxon = taxon
                special_taxon = True

            if special_taxon:
                taxon_name = taxon['commonNameEnglish'].strip().lower()
                taxon['taxon_name'] = taxon_name
                wi_taxon_name_to_taxa[taxon_name].append(taxon)
                continue

        # Do we have a species name?
        if not is_empty_wi_item(taxon['species']):

            # If 'species' is populated, 'genus' should always be populated; one item currently breaks
            # this rule.
            assert not is_empty_wi_item(taxon['genus'])
            
            taxon_name = (taxon['genus'].strip() + ' ' +
                        taxon['species'].strip()).strip().lower()
            assert not is_empty_wi_item(taxon['class']) and \
                not is_empty_wi_item(taxon['order']) and \
                not is_empty_wi_item(taxon['family'])

        elif not is_empty_wi_item(taxon['genus']):

            assert not is_empty_wi_item(taxon['class']) and \
                not is_empty_wi_item(taxon['order']) and \
                not is_empty_wi_item(taxon['family'])
            taxon_name = taxon['genus'].strip().lower()

        elif not is_empty_wi_item(taxon['family']):

            assert not is_empty_wi_item(taxon['class']) and \
                not is_empty_wi_item(taxon['order'])
            taxon_name = taxon['family'].strip().lower()

        elif not is_empty_wi_item(taxon['order']):

            assert not is_empty_wi_item(taxon['class'])
            taxon_name = taxon['order'].strip().lower()

        elif not is_empty_wi_item(taxon['class']):

            taxon_name = taxon['class'].strip().lower()

        if taxon_name is not None:
            assert taxon['taxonomyType'] == 'biological'
        else:
            assert taxon['taxonomyType'] == 'object'
            taxon_name = taxon['commonNameEnglish'].strip().lower()

        if taxon_name in wi_taxon_name_to_taxa:
            if taxon[id_column] in known_problematic_taxon_ids:
                print('Skipping problematic taxon ID {}'.format(taxon[id_column]))
            else:
                previous_taxa = wi_taxon_name_to_taxa[taxon_name]
                for previous_taxon in previous_taxa:
                    for level in ['class', 'order', 'family', 'genus', 'species']:
                        error_string = 'Error: taxon {} appeared previously in {} {} (as {}), now in {} {}'.format(
                            taxon_name,
                            level,previous_taxon[level],
                            previous_taxon['taxon_name'],
                            level,taxon[level])
                        assert taxonomy_items_equal(previous_taxon[level], taxon[level]), error_string
                    
        taxon['taxon_name'] = taxon_name
        if taxon_name == 'homo sapiens':
            human_taxa.append(taxon)
        wi_taxon_name_to_taxa[taxon_name].append(taxon)

    # ...for each taxon

    assert unknown_taxon is not None
    assert animal_taxon is not None
    assert blank_taxon is not None


    #%% Find redundant taxa

    taxon_names_with_multiple_entries = []
    for wi_taxon_name in wi_taxon_name_to_taxa:
        if len(wi_taxon_name_to_taxa[wi_taxon_name]) > 1:
            taxon_names_with_multiple_entries.append(wi_taxon_name)

    print('{} names have multiple entries\n:'.format(len(taxon_names_with_multiple_entries)))

    for s in taxon_names_with_multiple_entries:
        print(s)

    if False:
        pass

        #%% Manual review of redundant taxa
        
        s = taxon_names_with_multiple_entries[15]
        taxa = wi_taxon_name_to_taxa[s]
        for t in taxa:
            for k in t.keys():
                print('{}: {}'.format(k,t[k]))
            print()
            # print(t,end='\n\n')


    #%% Clean up redundant taxa

    taxon_name_to_preferred_taxon_id = {}

    # "helmeted guineafowl" vs "domestic guineafowl"
    taxon_name_to_preferred_taxon_id['numida meleagris'] = '83133617-8358-4910-82ee-4c23e40ba3dc' # 2005826 

    # "domestic turkey" vs. "wild turkey"
    taxon_name_to_preferred_taxon_id['meleagris gallopavo'] = 'c10547c3-1748-48bf-a451-8066c820f22f' # 2021598 

    # multiple sensible human entries
    taxon_name_to_preferred_taxon_id['homo sapiens'] = '990ae9dd-7a59-4344-afcb-1b7b21368000' # 2002045 

    # "domestic dog" and "dog-on-leash"
    taxon_name_to_preferred_taxon_id['canis familiaris'] = '3d80f1d6-b1df-4966-9ff4-94053c7a902a' # 2021548 

    # "small mammal" vs. "mammal"
    taxon_name_to_preferred_taxon_id['mammalia'] = 'f2d233e3-80e3-433d-9687-e29ecc7a467a' # 2021108 

    # "Hispaniolan Mango" vs. NaN
    taxon_name_to_preferred_taxon_id['anthracothorax dominicus'] = 'f94e6d97-59cf-4d38-a05a-a75efdd2863b'

    # "millipedes" vs. "Millipede"
    taxon_name_to_preferred_taxon_id['diplopoda'] =  '065884eb-4e64-4233-84dc-de25bd06ffd2' # 2021760

    # Different suborders: Squamata vs. Lacertilia
    taxon_name_to_preferred_taxon_id['squamata'] = '710c4066-bd5d-4313-bcf4-0217c4c84da7' # 2021703

    # Redundancy (both "beautiful firetail")
    taxon_name_to_preferred_taxon_id['stagonopleura bella'] = '7fec8e7e-fd3b-4d7f-99fd-3ade6f3bbaa5' # 2021939

    # "yellow wagtail" vs. "yellow crowned-wagtail"
    taxon_name_to_preferred_taxon_id['motacilla flava'] = 'ac6669bc-9f9e-4473-b609-b9082f9bf50c' # 2016194 

    # "dremomys species" vs. "dremomys genus"
    taxon_name_to_preferred_taxon_id['dremomys'] = '1507d153-af11-46f1-bfb8-77918d035ab3' # 2019370

    # "elk" vs. "domestic elk"
    taxon_name_to_preferred_taxon_id['cervus canadensis'] = 'c5ce946f-8f0d-4379-992b-cc0982381f5e' 

    # "American bison" vs. "domestic bison"
    taxon_name_to_preferred_taxon_id['bison bison'] = '539ebd55-081b-429a-9ae6-5a6a0f6999d4' # 2021593 

    # "woodrat or rat or mouse species" vs. "mouse species"
    taxon_name_to_preferred_taxon_id['muridae'] = 'e7503287-468c-45af-a1bd-a17821bb62f2' # 2021642 

    # both "southern sand frog"
    taxon_name_to_preferred_taxon_id['tomopterna adiastola'] = 'a5dc63cb-41be-4090-84a7-b944b16dcee4' # 2021834

    # sericornis species vs. scrubwren species
    taxon_name_to_preferred_taxon_id['sericornis'] = 'ad82c0ac-df48-4028-bf71-d2b2f4bc4129' # 2021776

        
    # taxon_name = list(taxon_name_to_preferred_taxon_id.keys())[0]
    for taxon_name in taxon_name_to_preferred_taxon_id.keys():
        
        candidate_taxa = wi_taxon_name_to_taxa[taxon_name]
        
        # If we've gotten this far, we should be choosing from multiple taxa.
        #
        # This will become untrue if any of these are resolved later, at which point we should
        # remove them from taxon_name_to_preferred_id
        assert len(candidate_taxa) > 1, 'Only one taxon available for {}'.format(taxon_name)
        
        # Choose the preferred taxa
        selected_taxa = [t for t in candidate_taxa if t[id_column] == \
                        taxon_name_to_preferred_taxon_id[taxon_name]]
        assert len(selected_taxa) == 1
        wi_taxon_name_to_taxa[taxon_name] = selected_taxa

    wi_taxon_name_to_taxon = {}

    for taxon_name in wi_taxon_name_to_taxa.keys():
        taxa = wi_taxon_name_to_taxa[taxon_name]
        assert len(taxa) == 1
        wi_taxon_name_to_taxon[taxon_name] = taxa[0]


    #%% Read supplementary mappings

    with open(lila_to_wi_supplementary_mapping_file, 'r') as f:
        lines = f.readlines()

    supplementary_lila_query_to_wi_query = {}

    for line in lines:
        # Each line is [lila query],[WI taxon name],[notes]
        tokens = line.strip().split(',')
        assert len(tokens) == 3
        lila_query = tokens[0].strip().lower()
        wi_taxon_name = tokens[1].strip().lower()
        assert wi_taxon_name in wi_taxon_name_to_taxa
        supplementary_lila_query_to_wi_query[lila_query] = wi_taxon_name


    #%% Map LILA categories to WI categories

    mismatches = set()
    mismatches_with_common_mappings = set()
    supplementary_mappings = set()

    all_searches = set()

    # Must be ordered from kingdom --> species
    lila_taxonomy_levels = ['kingdom', 'phylum', 'subphylum', 'superclass', 'class', 'subclass',
                            'infraclass', 'superorder', 'order', 'suborder', 'infraorder',
                            'superfamily', 'family', 'subfamily', 'tribe', 'genus', 'species']

    unknown_queries = set(
        ['unidentifiable', 'other', 'unidentified', 'unknown', 'unclassifiable'])
    blank_queries = set(['empty'])
    animal_queries = set(['animalia'])

    lila_dataset_category_to_wi_taxon = {}

    # i_taxon = 0; taxon = lila_taxonomy[i_taxon]; print(taxon)
    for i_taxon, lila_taxon in enumerate(lila_taxonomy):

        query = None

        lila_dataset_category = lila_taxon['dataset_name'] + ':' + lila_taxon['query']
        
        # Go from kingdom --> species, choosing the lowest-level description as the query
        for level in lila_taxonomy_levels:
            if isinstance(lila_taxon[level], str):
                query = lila_taxon[level]
                all_searches.add(query)

        if query is None:
            # E.g., 'car'
            query = lila_taxon['query']

        wi_taxon = None

        if query in unknown_queries:

            wi_taxon = unknown_taxon

        elif query in blank_queries:

            wi_taxon = blank_taxon

        elif query in animal_queries:

            wi_taxon = animal_taxon

        elif query in wi_taxon_name_to_taxon:

            wi_taxon = wi_taxon_name_to_taxon[query]

        elif query in supplementary_lila_query_to_wi_query:

            wi_taxon = wi_taxon_name_to_taxon[supplementary_lila_query_to_wi_query[query]]
            supplementary_mappings.add(query)
            # print('Made a supplementary mapping from {} to {}'.format(query,wi_taxon['taxon_name']))

        else:

            # print('No match for {}'.format(query))
            lila_common_name = lila_taxon['common_name']

            if lila_common_name in wi_common_name_to_taxon:
                wi_taxon = wi_common_name_to_taxon[lila_common_name]
                wi_common_name = wi_taxon['commonNameEnglish']
                wi_taxon_name = wi_taxon['taxon_name']
                if False:
                    print('LILA common name {} maps to WI taxon {} ({})'.format(lila_common_name,
                                                                                wi_taxon_name,
                                                                                wi_common_name))
                mismatches_with_common_mappings.add(query)

            else:

                mismatches.add(query)

        lila_dataset_category_to_wi_taxon[lila_dataset_category] = wi_taxon

    # ...for each LILA taxon

    print('Of {} entities, there are {} mismatches ({} mapped by common name) ({} mapped by supplementary mapping file)'.format(
        len(all_searches), len(mismatches), len(mismatches_with_common_mappings), len(supplementary_mappings)))

    assert len(mismatches) == 0


    #%% Manual mapping

    if not os.path.isfile(lila_to_wi_supplementary_mapping_file):
        print('Creating mapping file {}'.format(
            lila_to_wi_supplementary_mapping_file))
        with open(lila_to_wi_supplementary_mapping_file, 'w') as f:
            for query in mismatches:
                f.write(query + ',' + '\n')
    else:
        print('{} exists, not re-writing'.format(lila_to_wi_supplementary_mapping_file))


    #%% Build a dictionary from LILA dataset names and categories to LILA taxa

    lila_dataset_category_to_lila_taxon = {}

    # i_d = 0; d = lila_taxonomy[i_d]
    for i_d,d in enumerate(lila_taxonomy):
        lila_dataset_category = d['dataset_name'] + ':' + d['query']
        assert lila_dataset_category not in lila_dataset_category_to_lila_taxon
        lila_dataset_category_to_lila_taxon[lila_dataset_category] = d


    #%% Map LILA datasets to WI taxa, and count the number of each taxon available in each dataset

    with open(wi_mapping_table_file,'w') as f:
        
        f.write('lila_dataset_name,lila_category_name,wi_guid,wi_taxon_name,wi_common,count\n')
        
        # dataset_name = list(lila_dataset_to_categories.keys())[0]
        for dataset_name in lila_dataset_to_categories.keys():
            
            if '_bbox' in dataset_name:
                continue
            
            dataset_categories = lila_dataset_to_categories[dataset_name]
            
            # dataset_category = dataset_categories[0]
            for category in dataset_categories:
                
                lila_dataset_category = dataset_name + ':' + category['name'].strip().lower()
                if '#' in lila_dataset_category:
                    continue
                assert lila_dataset_category in lila_dataset_category_to_lila_taxon
                assert lila_dataset_category in lila_dataset_category_to_wi_taxon
                assert 'count' in category
        
                wi_taxon = lila_dataset_category_to_wi_taxon[lila_dataset_category]
                
                # Write out the dataset name, category name, WI GUID, WI scientific name, WI common name, 
                # and count
                s = f"{dataset_name},{category['name']},{wi_taxon['uniqueIdentifier']},"+\
                    f"{wi_taxon['taxon_name']},{wi_taxon['commonNameEnglish']},{category['count']}\n"
                f.write(s)
                
            # ...for each category in this dataset
                
        # ...for each dataset    

    # ...with open()



"""

 cct_field_adjustments.py

 CCT metadata was posted with int locations instead of strings.

 This script fixes those issues and rev's the version number.

"""

#%% Constants and environment

import json
import os

from megadetector.data_management.databases import integrity_check_json_db

inputJsonFile = r"D:\temp\CaltechCameraTraps_v2.0.json"
outputJsonFile = r"D:\temp\CaltechCameraTraps_v2.1.json"

assert os.path.isfile(inputJsonFile)


#%% Read .json file

with open(inputJsonFile,'r') as f:
    data = json.load(f)

images = data['images']
annotations = data['annotations']
categories = data['categories']
info = data['info']

print('Finished reading input .json')


#%% Rev version number, update field names and types

assert(info['version'] == 'Caltech Camera Traps - v2')
info['version'] = 2.1
info['description'] = 'Caltech Camera Traps: camera trap images collected from the NPS and the USGS with help from Justin Brown and Erin Boydston'

for image in images:
    
    assert 'location' in image and isinstance(image['location'],int)
    image['location'] = str(image['location'])
    

#%% Write json file
    
json.dump(data, open(outputJsonFile, 'w'), indent=4)

print('Finished writing output .json to {}'.format(outputJsonFile))


#%% Check output data file

integrity_check_json_db.integrity_check_json_db(outputJsonFile)



"""

annotation_constants.py

Defines default categories for MegaDetector output boxes.

Used throughout the repo; do not change unless you are Dan or Siyu.  In fact, do not change unless 
you are both Dan *and* Siyu.

We use integer IDs here; this is different from the MD .json file format,
where indices are string integers.

"""

#%% Constants

# MegaDetector output categories (the "empty" category is implicit)
detector_bbox_categories = [
    {'id': 0, 'name': 'empty'},
    {'id': 1, 'name': 'animal'},
    {'id': 2, 'name': 'person'},
    {'id': 3, 'name': 'vehicle'}
]

# This is used for choosing colors, so it ignores the "empty" class.
NUM_DETECTOR_CATEGORIES = len(detector_bbox_categories) - 1

detector_bbox_category_id_to_name = {}
detector_bbox_category_name_to_id = {}

for cat in detector_bbox_categories:
    detector_bbox_category_id_to_name[cat['id']] = cat['name']
    detector_bbox_category_name_to_id[cat['name']] = cat['id']




#
# make_full_SS_json.py
#
# Create a COCO-camera-traps .json file for Snapshot Serengeti data from
# the original .csv files provided on Dryad.
#
# This was used for "version 1.0" of the public Snapshot Serengeti archive; it's no 
# longer used as of v2.0 (early 2020).  See snapshot_serengeti_lila.py for the updated
# Snapshot Safari preparation process.
#

#%% Imports and constants

import csv
import json
import uuid
import datetime

output_file = '/datadrive/snapshotserengeti/databases/SnapshotSerengeti_multiple_classes.json'
csv_file_name = '/datadrive/snapshotserengeti/databases/consensus_data.csv'
all_image_file = '/datadrive/snapshotserengeti/databases/all_images.csv'


#%% Read annotation .csv file, format into a dictionary mapping field names to data arrays

data = []
with open(csv_file_name,'r') as f:
    reader = csv.reader(f, dialect = 'excel')
    for row in reader:
        data.append(row)

data_fields = data[0]

data_dicts = {}
for event in data[1:]:
    if event[0] not in data_dicts:
        data_dicts[event[0]] = []
    data_dicts[event[0]].append({data_fields[i]:event[i] for i in range(len(data_fields))})

# Count the number of images with multiple species
mult_species = 0
for event in data_dicts:
    if len(data_dicts[event]) > 1:
        mult_species += 1


#%% Read image .csv file, format into a dictionary mapping images to capture events

with open(all_image_file,'r') as f:
    reader = csv.reader(f,dialect = 'excel')
    next(reader)
    im_name_to_cap_id = {row[1]:row[0] for row in reader}
    
total_ims = len(im_name_to_cap_id)
total_seqs = len(data_dicts)
print('Percent seqs with mult species: ',float(mult_species)/float(total_seqs))


#%% Create CCT-style .json 

images = []
annotations = []
categories = []

capture_ims = {i:[] for i in im_name_to_cap_id.values()}
for im_id in im_name_to_cap_id:
    capture_ims[im_name_to_cap_id[im_id]].append(im_id)

im_to_seq_num = {im:None for im in im_name_to_cap_id}
for event in capture_ims:
    capture_ims[event] = sorted(capture_ims[event])
    seq_count = 0
    for im in capture_ims[event]:
        im_to_seq_num[im] = seq_count
        seq_count += 1

cat_to_id = {}
cat_to_id['empty'] = 0
cat_count = 1
seasons = []

for im_id in im_name_to_cap_id:
    
    im = {}
    im['id'] = im_id.split('.')[0]
    im['file_name'] = im_id
    
    im['location'] = im_id.split('/')[1]
    im['season'] = im_id.split('/')[0]
    if im['season'] not in seasons:
        seasons.append(im['season'])
    im['seq_id'] = im_name_to_cap_id[im_id]
    im['frame_num'] = im_to_seq_num[im_id]
    im['seq_num_frames'] = len(capture_ims[im['seq_id']])

    ann = {}
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']

    if im_name_to_cap_id[im_id] in data_dicts:
        im_data_per_ann = data_dicts[im_name_to_cap_id[im_id]]
        for im_data in im_data_per_ann:
            im['datetime'] = im_data['DateTime']
            if im_data['Species'] not in cat_to_id:
                cat_to_id[im_data['Species']] = cat_count
                cat_count += 1
            ann = {}
            ann['id'] = str(uuid.uuid1())
            ann['image_id'] = im['id']
            ann['category_id'] = cat_to_id[im_data['Species']]
            annotations.append(ann)
    else:
        ann = {}
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']
        ann['category_id'] = 0
        annotations.append(ann)
            
    # still need image width and height
    images.append(im)
    
# ...for each image
    
print(seasons)

for cat in cat_to_id:
    new_cat = {}
    new_cat['id'] = cat_to_id[cat]
    new_cat['name'] = cat
    categories.append(new_cat) 


#%% Write output files
    
json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
info = {}
info['year'] = 2018
info['version'] = 1
info['description'] = 'COCO style Snapshot Serengeti database'
info['contributor'] = 'SMB'
info['date_created'] = str(datetime.date.today())
json_data['info'] = info
json.dump(json_data,open(output_file,'w'))

print(images[0])
print(annotations[0]) 




#
# eMammal_helpers.py
#
# Support functions for processing eMammal metadata
#

#%% Constants and imports

from PIL import Image, ImageDraw
import matplotlib.pyplot as plt
import operator
from datetime import datetime


#%% Support functions

def clean_species_name(common_name):
    """
    Converts various forms of "human" to the token "human", and various forms
    of "empty" to the token "empty"
    """
    
    _people_tags = {
        'Bicycle',
        'Calibration Photos',
        'Camera Trapper',
        'camera trappper',
        'camera  trapper',
        'Homo sapien',
        'Homo sapiens',
        'Human, non staff',
        'Human, non-staff',
        'camera trappe',
        'Human non-staff',
        'Setup Pickup',
        'Vehicle'
    }
    PEOPLE_TAGS = {x.lower() for x in _people_tags}

    _no_animal_tags = {'No Animal', 'no  animal', 'Time Lapse', 'Camera Misfire', 'False trigger', 'Blank'}
    NO_ANIMAL_TAGS = {x.lower() for x in _no_animal_tags}

    common_name = common_name.lower().strip()
    if common_name in PEOPLE_TAGS:
        return 'human'

    if common_name in NO_ANIMAL_TAGS:
        return 'empty'

    return common_name


def clean_frame_number(img_frame):
    
    # pad to a total of 3 digits if < 1000, or 4 digits otherwise
    # img_frame is a string from the xml tree
    length = len(img_frame)

    assert length > 0
    assert length < 5

    # length 4 frame order is returned as is, others are left padded to be 3 digit long
    # we need to make sure img_frame has length 3 when it's < 1000 so we can match it to the iMerit labels
    if length == 1:
        return '00' + img_frame
    elif length == 2:
        return '0' + img_frame
    else:  # for '100' and '1000'
        return img_frame


def clean_frame_number_4_digit(img_frame):
    
    # pad to a total of 4 digits
    # img_frame is a string from the xml tree
    length = len(img_frame)

    assert length > 0
    assert length < 5

    # length 4 frame order is returned as is, others are left padded to be 3 digit long
    # we need to make sure img_frame has length 3 when it's < 1000 so we can match it to the iMerit labels
    if length == 1:
        return '000' + img_frame
    elif length == 2:
        return '00' + img_frame
    elif length == 3:
        return '0' + img_frame
    else:  # for'1000'
        return img_frame


def get_img_size(img_path):
    """
    There are ways to size the image without loading it into memory by reading its headers
    (https://github.com/scardine/image_size), but seems less reliable.

    Returns (-1, -1) if PIL could not open the image
    """

    try:
        im = Image.open(img_path)
        width, height = im.size
    except:
        return (-1, -1)
    return (width, height)


def get_total_from_distribution(d):
    
    total = 0
    for key, count in d.items():
        total += int(key) * count
    return total


def sort_dict_val_desc(d, percent=False):    
    """ Sort a dictionary by the values in descending order. Returns a list of tuples. """
    
    sorted_d = sorted(d.items(), key=operator.itemgetter(1), reverse=True)

    if percent:
        with_percent = []
        total = sum([t[1] for t in sorted_d])
        for k, v in sorted_d:
            p = '{:.1f}%'.format(100 * float(v) / total)
            with_percent.append((k, v, p))
        return with_percent

    return sorted_d


def plot_distribution(d, title='', top=15):
    
    if top is None or top > len(d):
        top = len(d)

    sorted_d = sort_dict_val_desc(d)

    top_d = sorted_d[:top]
    x = [t[0] for t in top_d]
    y = [t[1] for t in top_d]

    # others column
    others_d = sorted_d[top:]
    others_sum = sum([t[1] for t in others_d])
    x.append('others')
    y.append(others_sum)

    plt.bar(range(len(x)), y, align='center', facecolor='#57BC90', edgecolor=None)
    plt.xticks(range(len(x)), x, rotation=90)
    plt.title(title)
    plt.show()


def plot_histogram(l, title='', max_val=None, bins='auto'):
    
    if max_val:
        l = [x for x in l if x < max_val]

    plt.hist(l, bins=bins, facecolor='#57BC90', edgecolor=None)
    plt.title(title)
    plt.show()


def draw_bboxes(image, bboxes, classes, thickness=4, show_label=False):    
    """
    Draw bounding boxes on top of an image
    Args:
        image : Path to image or a loaded PIL image
        bboxes: A list of bboxes to draw on the image, each bbox is [top left x, top left y, width, height] in relative coordinates
        classes: A list of classes corresponding to the bboxes
        thickness: Thickness of the line to draw, minimum is 1
    Outputs:
        Image object with the bboxes and class labels annotated
    """
    if type(image) is str:
        img = Image.open(image)
    else:
        img = image.copy()

    draw = ImageDraw.Draw(img)
    img_width, img_height = img.size

    for i in range(len(bboxes)):
        x_rel, y_rel, w_rel, h_rel = bboxes[i]
        x = x_rel * img_width  # x and y are the top left
        y = y_rel * img_height
        w = w_rel * img_width
        h = h_rel * img_height

        if show_label:
            c = classes[i]
            draw.text((x + 15, y + 15), str(c), fill=(255, 0, 0, 255))

        for j in range(thickness):
            draw.rectangle(((x + j, y + j), (x + w + j, y + h + j)), outline='red')

    return img


def is_daytime(date_time):
    """ 
    Returns True if daytime as determined by the input timestamp, a rough 
    decision based on two seasons
    """
    
    # summer day hours: 6am - 7pm
    # others day hours: 7am - 6pm

    is_summer = True if date_time.month in [5, 6, 7, 8, 9] else False
    if is_summer:
        if date_time.hour >= 6 and date_time.hour <= 19:
            return True
        else:
            return False
    else:
        if date_time.hour >= 7 and date_time.hour <= 18:
            return True
        else:
            return False


def parse_timestamp(time_str):    
    """
    There are three datetime string formats in eMammal, and some have an empty field.
    Args:
        time_str: text in the tag ImageDateTime

    Returns:
        datetime object, error (None if no error)
    """

    if time_str == '' or time_str is None:
        return '', 'empty or None'
    try:
        res = datetime.strptime(time_str, '%Y-%m-%dT%H:%M:%S')
        return res, None
    except Exception:
        try:
            res = datetime.strptime(time_str, '%Y-%m-%d %H:%M:%S')
            return res, None
        except Exception:
            try:
                res = datetime.strptime(time_str, '%m/%d/%Y %H:%M')
                return res, None
            except:
                print('WARNING, time_str cannot be parsed {}.'.format(time_str))
                return time_str, 'cannot be parsed {}'.format(time_str)  # return original string
